"""
Command line interface programs for the GA4GH reference implementation.

TODO: document how to use these for development and simple deployment.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import logging
import unittest
import unittest.loader
import unittest.suite

import requests

import ga4gh.backend as backend
import ga4gh.client as client
import ga4gh.converters as converters
import ga4gh.frontend as frontend
import ga4gh.configtest as configtest
import ga4gh.exceptions as exceptions


# the maximum value of a long type in avro = 2**63 - 1
# (64 bit signed integer)
# http://avro.apache.org/docs/1.7.7/spec.html#schema_primitive
# AVRO_LONG_MAX = (1 << 63) - 1
# TODO in the meantime, this is the max value pysam can handle
# This should be removed once pysam input sanitisation has been
# implemented.
AVRO_LONG_MAX = 2**31 - 1


##############################################################################
# Server
##############################################################################


def addServerOptions(parser):
    parser.add_argument(
        "--port", "-P", default=8000, type=int,
        help="The port to listen on")
    parser.add_argument(
        "--host", "-H", default="127.0.0.1",
        help="The server host string; use 0.0.0.0 to allow all connections.")
    parser.add_argument(
        "--config", "-c", default='DevelopmentConfig', type=str,
        help="The configuration to use")
    parser.add_argument(
        "--config-file", "-f", type=str, default=None,
        help="The configuration file to use")
    parser.add_argument(
        "--tls", "-t", action="store_true", default=False,
        help="Start in TLS (https) mode.")
    parser.add_argument(
        "--dont-use-reloader", default=False, action="store_true",
        help="Don't use the flask reloader")
    addDisableUrllibWarningsArgument(parser)


def server_main(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(
            description="GA4GH reference server")
    addServerOptions(parser)
    args = parser.parse_args()
    if args.disable_urllib_warnings:
        requests.packages.urllib3.disable_warnings()
    frontend.configure(
        args.config_file, args.config, args.port)
    sslContext = None
    if args.tls or ("OIDC_PROVIDER" in frontend.app.config):
        sslContext = "adhoc"
    frontend.app.run(
        host=args.host, port=args.port,
        use_reloader=not args.dont_use_reloader, ssl_context=sslContext)


##############################################################################
# Client
##############################################################################


def verbosityToLogLevel(verbosity):
    """
    Returns the specfied verbosity level interpreted as a logging level.
    """
    ret = 0
    if verbosity == 1:
        ret = logging.INFO
    elif verbosity >= 2:
        ret = logging.DEBUG
    return ret


class AbstractQueryRunner(object):
    """
    Abstract base class for runner classes
    """
    def __init__(self, args):
        self._key = args.key
        # TODO this is an experimental addition which is useful for
        # testing. We should think about this and document it if we
        # this it's a useful feature. There is an argument for pushing
        # the backend instantiation into the client, and letting the
        # client be a factory, instantiating the correct Client class
        # depending on the prefix.
        filePrefix = "file://"
        if args.baseUrl.startswith(filePrefix):
            dataDir = args.baseUrl[len(filePrefix):]
            theBackend = backend.FileSystemBackend(dataDir)
            self._client = client.LocalClient(theBackend)
        else:
            self._client = client.HttpClient(
                args.baseUrl, verbosityToLogLevel(args.verbose), self._key)


class FormattedOutputRunner(AbstractQueryRunner):
    """
    Superclass of runners that support output in common formats.
    """
    def __init__(self, args):
        super(FormattedOutputRunner, self).__init__(args)
        self._output = self._textOutput
        if args.outputFormat == "json":
            self._output = self._jsonOutput

    def _jsonOutput(self, gaObjects):
        """
        Outputs the specified protocol objects as one JSON string per
        line.
        """
        for gaObject in gaObjects:
            print(gaObject.toJsonString())

    def _textOutput(self, gaObjects):
        """
        Outputs a text summary of the specified protocol objects, one
        per line.
        """
        for gaObject in gaObjects:
            print(gaObject.id, gaObject.name, sep="\t")


class AbstractGetRunner(FormattedOutputRunner):
    """
    Abstract base class for get runner classes
    """
    def __init__(self, args):
        super(AbstractGetRunner, self).__init__(args)
        self._id = args.id

    def run(self):
        response = self._method(self._id)
        self._output([response])


class AbstractSearchRunner(FormattedOutputRunner):
    """
    Abstract base class for search runner classes
    """

    def __init__(self, args):
        super(AbstractSearchRunner, self).__init__(args)
        self._pageSize = args.pageSize
        self._client.setPageSize(self._pageSize)

    def getAllDatasets(self):
        """
        Returns all datasets on the server.
        """
        return self._client.searchDatasets()

    def getAllVariantSets(self):
        """
        Returns all variant sets on the server.
        """
        for dataset in self.getAllDatasets():
            iterator = self._client.searchVariantSets(datasetId=dataset.id)
            for variantSet in iterator:
                yield variantSet

    def getAllReadGroupSets(self):
        """
        Returns all readgroup sets on the server.
        """
        for dataset in self.getAllDatasets():
            iterator = self._client.searchReadGroupSets(
                datasetId=dataset.id)
            for readGroupSet in iterator:
                yield readGroupSet

    def getAllReadGroups(self):
        """
        Get all read groups in a read group set
        """
        for dataset in self.getAllDatasets():
            iterator = self._client.searchReadGroupSets(
                datasetId=dataset.id)
            for readGroupSet in iterator:
                readGroupSet = self._client.getReadGroupSet(readGroupSet.id)
                for readGroup in readGroupSet.readGroups:
                    yield readGroup.id

    def getAllReferenceSets(self):
        """
        Returns all reference sets on the server.
        """
        return self._client.searchReferenceSets()


# Runners for the various search methods

class SearchDatasetsRunner(AbstractSearchRunner):
    """
    Runner class for the datasets/search method
    """
    def __init__(self, args):
        super(SearchDatasetsRunner, self).__init__(args)

    def run(self):
        iterator = self._client.searchDatasets()
        self._output(iterator)


class SearchReferenceSetsRunner(AbstractSearchRunner):
    """
    Runner class for the referencesets/search method.
    """
    def __init__(self, args):
        super(SearchReferenceSetsRunner, self).__init__(args)
        self._accession = args.accession
        self._md5checksum = args.md5checksum

    def run(self):
        iterator = self._client.searchReferenceSets(
            accession=self._accession, md5checksum=self._md5checksum)
        self._output(iterator)


class SearchReferencesRunner(AbstractSearchRunner):
    """
    Runner class for the references/search method
    """
    def __init__(self, args):
        super(SearchReferencesRunner, self).__init__(args)
        self._referenceSetId = args.referenceSetId
        self._accession = args.accession
        self._md5checksum = args.md5checksum

    def _run(self, referenceSetId):
        iterator = self._client.searchReferences(
            accession=self._accession, md5checksum=self._md5checksum,
            referenceSetId=referenceSetId)
        self._output(iterator)

    def run(self):
        if self._referenceSetId is None:
            for referenceSet in self.getAllReferenceSets():
                self._run(referenceSet.id)
        else:
            self._run(self._referenceSetId)


class SearchVariantSetsRunner(AbstractSearchRunner):
    """
    Runner class for the variantsets/search method.
    """
    def __init__(self, args):
        super(SearchVariantSetsRunner, self).__init__(args)
        self._datasetId = args.datasetId

    def _run(self, datasetId):
        iterator = self._client.searchVariantSets(datasetId=datasetId)
        self._output(iterator)

    def run(self):
        if self._datasetId is None:
            for dataset in self.getAllDatasets():
                self._run(dataset.id)
        else:
            self._run(self._datasetId)


class SearchReadGroupSetsRunner(AbstractSearchRunner):
    """
    Runner class for the readgroupsets/search method
    """
    def __init__(self, args):
        super(SearchReadGroupSetsRunner, self).__init__(args)
        self._datasetId = args.datasetId
        self._name = args.name

    def _run(self, datasetId):
        iterator = self._client.searchReadGroupSets(
            datasetId=datasetId, name=self._name)
        self._output(iterator)

    def run(self):
        if self._datasetId is None:
            for dataset in self.getAllDatasets():
                self._run(dataset.id)
        else:
            self._run(self._datasetId)


class SearchCallSetsRunner(AbstractSearchRunner):
    """
    Runner class for the callsets/search method
    """
    def __init__(self, args):
        super(SearchCallSetsRunner, self).__init__(args)
        self._variantSetId = args.variantSetId
        self._name = args.name

    def _run(self, variantSetId):
        iterator = self._client.searchCallSets(
            variantSetId=variantSetId, name=self._name)
        self._output(iterator)

    def run(self):
        if self._variantSetId is None:
            for variantSet in self.getAllVariantSets():
                self._run(variantSet.id)
        else:
            self._run(self._variantSetId)


class VariantFormatterMixin(object):
    """
    Simple mixin to format variant objects.
    """
    def _textOutput(self, gaObjects):
        """
        Prints out the specified Variant objects in a VCF-like form.
        """
        for variant in gaObjects:
            print(
                variant.id, variant.variantSetId, variant.names,
                variant.referenceName, variant.start, variant.end,
                variant.referenceBases, variant.alternateBases,
                sep="\t", end="\t")
            for key, value in variant.info.items():
                print(key, value, sep="=", end=";")
            print("\t", end="")
            for c in variant.calls:
                print(
                    c.callSetId, c.genotype, c.genotypeLikelihood, c.info,
                    c.phaseset, sep=":", end="\t")
            print()


class SearchVariantsRunner(VariantFormatterMixin, AbstractSearchRunner):
    """
    Runner class for the variants/search method.
    """
    def __init__(self, args):
        super(SearchVariantsRunner, self).__init__(args)
        self._referenceName = args.referenceName
        self._variantSetId = args.variantSetId
        self._start = args.start
        self._end = args.end
        if args.callSetIds == []:
            self._callSetIds = []
        elif args.callSetIds == '*':
            self._callSetIds = None
        else:
            self._callSetIds = args.callSetIds.split(",")

    def _run(self, variantSetId):
        iterator = self._client.searchVariants(
            start=self._start, end=self._end,
            referenceName=self._referenceName,
            variantSetId=variantSetId, callSetIds=self._callSetIds)
        self._output(iterator)

    def run(self):
        if self._variantSetId is None:
            for variantSet in self.getAllVariantSets():
                self._run(variantSet.id)
        else:
            self._run(self._variantSetId)


class SearchReadsRunner(AbstractSearchRunner):
    """
    Runner class for the reads/search method
    """
    def __init__(self, args):
        super(SearchReadsRunner, self).__init__(args)
        self._start = args.start
        self._end = args.end
        self._referenceId = args.referenceId
        self._readGroupIds = None
        if args.readGroupIds is not None:
            self._readGroupIds = args.readGroupIds.split(",")

    def _run(self, referenceGroupId, referenceId=None):
        """
        automatically guess reference id if not passed
        """
        # check if we can get reference id from rg
        if referenceId is None:
            referenceId = self._referenceId
        if referenceId is None:
            rg = self._client.getReadGroup(readGroupId=referenceGroupId)
            iterator = self._client.searchReferences(rg.referenceSetId)
            for reference in iterator:
                self._run(referenceGroupId, reference.id)
        else:
            iterator = self._client.searchReads(
                readGroupIds=[referenceGroupId], referenceId=referenceId,
                start=self._start, end=self._end)
            self._output(iterator)

    def run(self):
        """
        Iterate passed read group ids, or go through all available read groups
        """
        if not self._readGroupIds:
            for referenceGroupId in self.getAllReadGroups():
                self._run(referenceGroupId)
        else:
            for referenceGroupId in self._readGroupIds:
                self._run(referenceGroupId)

    def _textOutput(self, gaObjects):
        """
        Prints out the specified Variant objects in a VCF-like form.
        """
        for read in gaObjects:
            # TODO add in some more useful output here.
            print(read.id)


# ListReferenceBases is an oddball, and doesn't fit either get or
# search patterns.

class ListReferenceBasesRunner(AbstractQueryRunner):
    """
    Runner class for the references/{id}/bases method
    """
    def __init__(self, args):
        super(ListReferenceBasesRunner, self).__init__(args)
        self._referenceId = args.id
        self._start = args.start
        self._end = args.end
        self._outputFormat = args.outputFormat

    def run(self):
        sequence = self._client.listReferenceBases(
            self._referenceId, self._start, self._end)
        if self._outputFormat == "text":
            print(sequence)
        else:
            start = self._start if self._start else ""
            end = self._end if self._end else ""
            print(">{}:{}-{}".format(self._referenceId, start, end))

            textWidth = 70
            for index in xrange(0, len(sequence), textWidth):
                print(sequence[index: index+textWidth])


# Runners for the various GET methods.

class GetReferenceSetRunner(AbstractGetRunner):
    """
    Runner class for the referencesets/{id} method
    """
    def __init__(self, args):
        super(GetReferenceSetRunner, self).__init__(args)
        self._method = self._client.getReferenceSet


class GetReferenceRunner(AbstractGetRunner):
    """
    Runner class for the references/{id} method
    """
    def __init__(self, args):
        super(GetReferenceRunner, self).__init__(args)
        self._method = self._client.getReference


class GetReadGroupSetRunner(AbstractGetRunner):
    """
    Runner class for the readgroupsets/{id} method
    """
    def __init__(self, args):
        super(GetReadGroupSetRunner, self).__init__(args)
        self._method = self._client.getReadGroupSet


class GetReadGroupRunner(AbstractGetRunner):
    """
    Runner class for the references/{id} method
    """
    def __init__(self, args):
        super(GetReadGroupRunner, self).__init__(args)
        self._method = self._client.getReadGroup


class GetCallsetRunner(AbstractGetRunner):
    """
    Runner class for the callsets/{id} method
    """
    def __init__(self, args):
        super(GetCallsetRunner, self).__init__(args)
        self._method = self._client.getCallset


class GetDatasetRunner(AbstractGetRunner):
    """
    Runner class for the datasets/{id} method
    """
    def __init__(self, args):
        super(GetDatasetRunner, self).__init__(args)
        self._method = self._client.getDataset


class GetVariantRunner(VariantFormatterMixin, AbstractGetRunner):
    """
    Runner class for the variants/{id} method
    """
    def __init__(self, args):
        super(GetVariantRunner, self).__init__(args)
        self._method = self._client.getVariant


def addDisableUrllibWarningsArgument(parser):
    parser.add_argument(
        "--disable-urllib-warnings", default=False, action="store_true",
        help="Disable urllib3 warnings")


def addVariantSearchOptions(parser):
    """
    Adds common options to a variant searches command line parser.
    """
    addVariantSetIdArgument(parser)
    addReferenceNameArgument(parser)
    addCallSetIdsArgument(parser)
    addStartArgument(parser)
    addEndArgument(parser)
    addPageSizeArgument(parser)


def addVariantSetIdArgument(parser):
    parser.add_argument(
        "--variantSetId", "-V", default=None,
        help="The variant set id to search over")


def addReferenceNameArgument(parser):
    parser.add_argument(
        "--referenceName", "-r", default="1",
        help="Only return variants on this reference.")


def addCallSetIdsArgument(parser):
    parser.add_argument(
        "--callSetIds", "-c", default=[],
        help="""Return variant calls which belong to call sets
            with these IDs. Pass in IDs as a comma separated list (no spaces).
            Use '*' to request all call sets (the quotes are important!).
            """)


def addStartArgument(parser):
    parser.add_argument(
        "--start", "-s", default=0, type=int,
        help="The start of the search range (inclusive).")


def addEndArgument(parser, defaultValue=AVRO_LONG_MAX):
    parser.add_argument(
        "--end", "-e", default=defaultValue, type=int,
        help="The end of the search range (exclusive).")


def addIdArgument(parser):
    parser.add_argument("id", default=None, help="The id of the object")


def addGetArguments(parser):
    addUrlArgument(parser)
    addIdArgument(parser)
    addOutputFormatArgument(parser)


def addUrlArgument(parser):
    """
    Adds the URL endpoint argument to the specified parser.
    """
    parser.add_argument("baseUrl", help="The URL of the API endpoint")


def addOutputFormatArgument(parser):
    parser.add_argument(
        "--outputFormat", "-O", choices=['text', 'json'], default="text",
        help=(
            "The format for object output. Currently supported are "
            "'text' (default), which gives a short summary of the object and "
            "'json', which outputs each object in line-delimited JSON"))


def addAccessionArgument(parser):
    parser.add_argument(
        "--accession", default=None,
        help="The accession to search for")


def addMd5ChecksumArgument(parser):
    parser.add_argument(
        "--md5checksum", default=None,
        help="The md5checksum to search for")


def addPageSizeArgument(parser):
    parser.add_argument(
        "--pageSize", "-m", default=None, type=int,
        help=(
            "The maximum number of results returned in one page. "
            "The default is to let the server decide how many "
            "results to return in a single page."))


def addDatasetIdArgument(parser):
    parser.add_argument(
        "--datasetId", default=None,
        help="The datasetId to search over")


def addReferenceSetIdArgument(parser):
    parser.add_argument(
        "--referenceSetId", default=None,
        help="The referenceSet to search over")


def addNameArgument(parser):
    parser.add_argument(
        "--name", default=None,
        help="The name to search over")


def addClientGlobalOptions(parser):
    parser.add_argument(
        '--verbose', '-v', action='count', default=0,
        help="Increase verbosity; can be supplied multiple times")
    parser.add_argument(
        "--key", "-k", default='invalid',
        help="Auth Key. Found on server index page.")
    addDisableUrllibWarningsArgument(parser)


def addHelpParser(subparsers):
    parser = subparsers.add_parser(
        "help", description="ga4gh_client help",
        help="show this help message and exit")
    return parser


def addVariantsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "variants-search",
        description="Search for variants",
        help="Search for variants.")
    parser.set_defaults(runner=SearchVariantsRunner)
    addUrlArgument(parser)
    addOutputFormatArgument(parser)
    addVariantSearchOptions(parser)
    return parser


def addVariantSetsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "variantsets-search",
        description="Search for variantSets",
        help="Search for variantSets.")
    parser.set_defaults(runner=SearchVariantSetsRunner)
    addOutputFormatArgument(parser)
    addUrlArgument(parser)
    addPageSizeArgument(parser)
    addDatasetIdArgument(parser)
    return parser


def addReferenceSetsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "referencesets-search",
        description="Search for referenceSets",
        help="Search for referenceSets")
    parser.set_defaults(runner=SearchReferenceSetsRunner)
    addUrlArgument(parser)
    addOutputFormatArgument(parser)
    addPageSizeArgument(parser)
    addAccessionArgument(parser)
    addMd5ChecksumArgument(parser)
    parser.add_argument(
        "--assemblyId",
        help="The assembly id to search for")
    return parser


def addReferencesSearchParser(subparsers):
    parser = subparsers.add_parser(
        "references-search",
        description="Search for references",
        help="Search for references")
    parser.set_defaults(runner=SearchReferencesRunner)
    addUrlArgument(parser)
    addOutputFormatArgument(parser)
    addPageSizeArgument(parser)
    addAccessionArgument(parser)
    addMd5ChecksumArgument(parser)
    addReferenceSetIdArgument(parser)
    return parser


def addReadGroupSetsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "readgroupsets-search",
        description="Search for readGroupSets",
        help="Search for readGroupSets")
    parser.set_defaults(runner=SearchReadGroupSetsRunner)
    addUrlArgument(parser)
    addOutputFormatArgument(parser)
    addPageSizeArgument(parser)
    addDatasetIdArgument(parser)
    addNameArgument(parser)
    return parser


def addCallsetsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "callsets-search",
        description="Search for callSets",
        help="Search for callSets")
    parser.set_defaults(runner=SearchCallSetsRunner)
    addUrlArgument(parser)
    addOutputFormatArgument(parser)
    addPageSizeArgument(parser)
    addNameArgument(parser)
    addVariantSetIdArgument(parser)
    return parser


def addReadsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "reads-search",
        description="Search for reads",
        help="Search for reads")
    parser.set_defaults(runner=SearchReadsRunner)
    addOutputFormatArgument(parser)
    addReadsSearchParserArguments(parser)
    return parser


def addDatasetsGetParser(subparsers):
    parser = subparsers.add_parser(
        "datasets-get",
        description="Get a dataset",
        help="Get a dataset")
    parser.set_defaults(runner=GetDatasetRunner)
    addGetArguments(parser)


def addDatasetsSearchParser(subparsers):
    parser = subparsers.add_parser(
        "datasets-search",
        description="Search for datasets",
        help="Search for datasets")
    parser.set_defaults(runner=SearchDatasetsRunner)
    addUrlArgument(parser)
    addPageSizeArgument(parser)
    addOutputFormatArgument(parser)
    return parser


def addReadsSearchParserArguments(parser):
    addUrlArgument(parser)
    addPageSizeArgument(parser)
    addStartArgument(parser)
    addEndArgument(parser)
    parser.add_argument(
        "--readGroupIds", default=None,
        help="The readGroupIds to search over")
    parser.add_argument(
        "--referenceId", default=None,
        help="The referenceId to search over")


def addReferenceSetsGetParser(subparsers):
    parser = subparsers.add_parser(
        "referencesets-get",
        description="Get a referenceset",
        help="Get a referenceset")
    parser.set_defaults(runner=GetReferenceSetRunner)
    addGetArguments(parser)


def addReferencesGetParser(subparsers):
    parser = subparsers.add_parser(
        "references-get",
        description="Get a reference",
        help="Get a reference")
    parser.set_defaults(runner=GetReferenceRunner)
    addGetArguments(parser)


def addReadGroupSetsGetParser(subparsers):
    parser = subparsers.add_parser(
        "readgroupsets-get",
        description="Get a read group set",
        help="Get a read group set")
    parser.set_defaults(runner=GetReadGroupSetRunner)
    addGetArguments(parser)


def addReadGroupsGetParser(subparsers):
    parser = subparsers.add_parser(
        "readgroups-get",
        description="Get a read group",
        help="Get a read group")
    parser.set_defaults(runner=GetReadGroupRunner)
    addGetArguments(parser)


def addCallsetsGetParser(subparsers):
    parser = subparsers.add_parser(
        "callsets-get",
        description="Get a callset",
        help="Get a callset")
    parser.set_defaults(runner=GetCallsetRunner)
    addGetArguments(parser)


def addVariantsGetParser(subparsers):
    parser = subparsers.add_parser(
        "variants-get",
        description="Get a variant",
        help="Get a variant")
    parser.set_defaults(runner=GetVariantRunner)
    addGetArguments(parser)


def addReferencesBasesListParser(subparsers):
    parser = subparsers.add_parser(
        "references-list-bases",
        description="List bases of a reference",
        help="List bases of a reference")
    parser.add_argument(
        "--outputFormat", "-O", choices=['text', 'fasta'], default="text",
        help=(
            "The format for sequence output. Currently supported are "
            "'text' (default), which prints the sequence out directly and "
            "'fasta', which formats the sequence into fixed width FASTA"))
    parser.set_defaults(runner=ListReferenceBasesRunner)
    addUrlArgument(parser)
    addIdArgument(parser)
    addStartArgument(parser)
    addEndArgument(parser, defaultValue=None)


def getClientParser():
    parser = argparse.ArgumentParser(
        description="GA4GH reference client")
    addClientGlobalOptions(parser)
    subparsers = parser.add_subparsers(title='subcommands',)
    addHelpParser(subparsers)
    addVariantsSearchParser(subparsers)
    addVariantSetsSearchParser(subparsers)
    addReferenceSetsSearchParser(subparsers)
    addReferencesSearchParser(subparsers)
    addReadGroupSetsSearchParser(subparsers)
    addCallsetsSearchParser(subparsers)
    addReadsSearchParser(subparsers)
    addDatasetsSearchParser(subparsers)
    addReferenceSetsGetParser(subparsers)
    addReferencesGetParser(subparsers)
    addReadGroupSetsGetParser(subparsers)
    addReadGroupsGetParser(subparsers)
    addCallsetsGetParser(subparsers)
    addVariantsGetParser(subparsers)
    addDatasetsGetParser(subparsers)
    addReferencesBasesListParser(subparsers)
    return parser


def client_main(args=None):
    parser = getClientParser()
    args = parser.parse_args(args)
    if "runner" not in args:
        parser.print_help()
    else:
        if args.disable_urllib_warnings:
            requests.packages.urllib3.disable_warnings()
        try:
            runner = args.runner(args)
            runner.run()
        except (exceptions.BaseClientException,
                requests.exceptions.RequestException) as exception:
            # TODO suppress exception unless debug settings are enabled
            raise exception


##############################################################################
# ga2vcf
##############################################################################

class Ga2VcfRunner(SearchVariantsRunner):
    """
    Runner class for the ga2vcf
    """
    def __init__(self, args):
        super(Ga2VcfRunner, self).__init__(args)
        self._outputFile = args.outputFile
        self._binaryOutput = False
        if args.outputFormat == "bcf":
            self._binaryOutput = True

    def run(self):
        variantSet = self._client.getVariantSet(self._variantSetId)
        iterator = self._client.searchVariants(
            start=self._start, end=self._end,
            referenceName=self._referenceName,
            variantSetId=self._variantSetId,
            callSetIds=self._callSetIds)
        # do conversion
        vcfConverter = converters.VcfConverter(
            variantSet, iterator, self._outputFile, self._binaryOutput)
        vcfConverter.convert()


def addOutputFileArgument(parser):
    parser.add_argument(
        "--outputFile", "-o", default=None,
        help="the file to write the output to")


def getGa2VcfParser():
    parser = argparse.ArgumentParser(
        description=(
            "GA4GH VCF conversion tool. Converts variant information "
            "stored in a GA4GH repository into VCF format."))
    addClientGlobalOptions(parser)
    addOutputFileArgument(parser)
    addUrlArgument(parser)
    parser.add_argument("variantSetId", help="The variant set to convert")
    parser.add_argument(
        "--outputFormat", "-O", choices=['vcf', 'bcf'], default="vcf",
        help=(
            "The format for object output. Currently supported are "
            "'vcf' (default), which is a text-based format and "
            "'bcf', which is the binary equivalent"))
    addReferenceNameArgument(parser)
    addCallSetIdsArgument(parser)
    addStartArgument(parser)
    addEndArgument(parser)
    addPageSizeArgument(parser)
    return parser


def ga2vcf_main():
    parser = getGa2VcfParser()
    args = parser.parse_args()
    if "baseUrl" not in args:
        parser.print_help()
    else:
        runner = Ga2VcfRunner(args)
        runner.run()


##############################################################################
# ga2sam
##############################################################################


class Ga2SamRunner(SearchReadsRunner):
    """
    Runner class for the ga2vcf
    """
    def __init__(self, args):
        args.readGroupIds = args.readGroupId
        super(Ga2SamRunner, self).__init__(args)
        self._outputFile = args.outputFile
        self._binaryOutput = False
        if args.outputFormat == "bam":
            self._binaryOutput = True

    def run(self):
        samConverter = converters.SamConverter(
            self._client, readGroupId=self._readGroupIds[0],
            referenceId=self._referenceId, start=self._start, end=self._end,
            outputFileName=self._outputFile, binaryOutput=self._binaryOutput)
        samConverter.convert()


def getGa2SamParser():
    parser = argparse.ArgumentParser(
        description="GA4GH SAM conversion tool")
    addClientGlobalOptions(parser)
    addUrlArgument(parser)
    parser.add_argument(
        "readGroupId",
        help="The ReadGroup to convert to SAM/BAM format.")
    addPageSizeArgument(parser)
    addStartArgument(parser)
    addEndArgument(parser)
    parser.add_argument(
        "--referenceId", default=None,
        help="The referenceId to search over")
    parser.add_argument(
        "--outputFormat", "-O", default="sam", choices=["sam", "bam"],
        help=(
            "The format for object output. Currently supported are "
            "'sam' (default), which is a text-based format and "
            "'bam', which is the binary equivalent"))
    addOutputFileArgument(parser)
    return parser


def ga2sam_main():
    parser = getGa2SamParser()
    args = parser.parse_args()
    if "baseUrl" not in args:
        parser.print_help()
    else:
        runner = Ga2SamRunner(args)
        runner.run()


##############################################################################
# Configuration testing
##############################################################################


class SimplerResult(unittest.TestResult):
    """
    The TestResult class gives formatted tracebacks as error messages, which
    is not what we want. Instead we just want the error message from the
    err praram. Hence this subclass.
    """
    def addError(self, test, err):
        self.errors.append((test,
                            "{0}: {1}".format(err[0].__name__, err[1])))

    def addFailure(self, test, err):
        self.failures.append((test,
                              "{0}: {1}".format(err[0].__name__, err[1])))


def configtest_main(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser(
            description="GA4GH server configuration validator")
    parser.add_argument(
        "--config", "-c", default='DevelopmentConfig', type=str,
        help="The configuration to use")
    parser.add_argument(
        "--config-file", "-f", type=str, default=None,
        help="The configuration file to use")
    args = parser.parse_args()
    configStr = 'ga4gh.serverconfig:{0}'.format(args.config)

    configtest.TestConfig.configStr = configStr
    configtest.TestConfig.configFile = args.config_file
    configtest.TestConfig.configEnv = "GA4GH_CONFIGURATION"

    loader = unittest.TestLoader()
    tests = loader.loadTestsFromModule(configtest)
    results = SimplerResult()
    tests.run(results)

    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger(__name__)
    log.info('{0} Tests run. {1} errors, {2} failures, {3} skipped'.
             format(results.testsRun,
                    len(results.errors),
                    len(results.failures),
                    len(results.skipped)))
    for result in results.errors:
        if result is not None:
            log.critical('Error: {0}: {1}'.format(result[0].id(), result[1]))
    for result in results.failures:
        if result is not None:
            log.critical('Failure: {0}: {1}'.format(result[0].id(), result[1]))
    for result in results.skipped:
        if result is not None:
            log.info('Skipped: {0}: {1}'.format(result[0].id(), result[1]))



"""
A script that takes the compliance dataset (the released version
of which is at https://github.com/ga4gh/compliance/tree/master/test-data)
and turns it into a directory bundle of binary and JSON files suitable
for use by the reference server.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import shutil
import argparse
import os
import sys
import glob
import pysam
import utils


class ComplianceDataMunger(object):
    def __init__(self, args):
        self.inputDirectory = args.inputDirectory
        self.outputDirectory = args.outputDirectory

        # get all the reference files (they'll be the ones with .fa extension)
        self.referenceFiles = map(
            os.path.basename, glob.glob(
                os.path.join(self.inputDirectory, "*.fa")))

        self.refsetsDirectory = os.path.join(
            self.outputDirectory, "referenceSets")
        self.hg37Directory = os.path.join(self.refsetsDirectory, "hg37")

        # datasets
        self.datasetsDirectory = os.path.join(self.outputDirectory, "datasets")
        self.readFiles = map(
            os.path.basename, glob.glob(
                os.path.join(self.inputDirectory, "*.sam")))
        self.variantFiles = map(
            os.path.basename, glob.glob(
                os.path.join(self.inputDirectory, "*.vcf")))

        self.datasets = [d for d in set([p.split('_')[0] for p in
                                         self.readFiles + self.variantFiles])]
        self.datasetReads = dict()
        self.datasetVariants = dict()
        for ds in self.datasets:
            self.datasetReads[ds] = [r for r in
                                     self.readFiles if r.startswith(ds)]

        # Variants themselves are split into groups,
        # based on second part of the _ split:
        for ds in self.datasets:
            self.datasetVariants[ds] = dict()
            # only those variants inside this dataset
            dsvlist = [v for v in self.variantFiles if v.startswith(ds)]
            # create nested dictionary based on group belonging
            for dsv in dsvlist:
                dsvGroup = dsv.split('_')[1]
                self.datasetVariants[ds][dsvGroup] = \
                    self.datasetVariants[ds].get(dsvGroup, []) + [dsv]

        self.datasetDirs = [os.path.join(self.outputDirectory, ds)
                            for ds in self.datasets]

    def run(self):
        if not os.path.exists(self.outputDirectory):
            os.makedirs(self.outputDirectory)

        # Clean out, make and re-populate references directory
        # For now, assume a single, statically-named referenceSet
        print("Converting references...", file=sys.stderr)
        shutil.rmtree(self.refsetsDirectory, ignore_errors=True)
        os.makedirs(self.refsetsDirectory)
        shutil.copy(
            os.path.join(self.inputDirectory, "referenceset_hg37.json"),
            os.path.join(self.refsetsDirectory, "hg37.json"))

        os.makedirs(self.hg37Directory)
        for refFile in self.referenceFiles:
            refBase = os.path.splitext(refFile)[0]
            destFastaFilename = os.path.join(
                self.hg37Directory, refBase) + ".fa"
            shutil.copy(os.path.join(self.inputDirectory, refBase) + ".fa",
                        destFastaFilename)
            pysam.tabix_compress(destFastaFilename, destFastaFilename + ".gz")
            refFasta = pysam.FastaFile(destFastaFilename + ".gz")
            refFasta.close()
            os.remove(destFastaFilename)
            shutil.copy(
                os.path.join(self.inputDirectory, refBase) + ".json",
                os.path.join(self.hg37Directory, refBase) + ".json")

        # Clean out, make and repopulate dataset directories
        shutil.rmtree(self.datasetsDirectory, ignore_errors=True)
        os.makedirs(self.datasetsDirectory)

        for ds in self.datasets:
            dsdir = os.path.join(self.datasetsDirectory, ds)
            os.makedirs(dsdir)

            # Reads
            print("Converting reads...", file=sys.stderr)
            dsReadsdir = os.path.join(dsdir, "reads")
            os.makedirs(dsReadsdir)
            for readFile in self.datasetReads[ds]:
                destFile = os.path.join(
                    dsReadsdir,
                    readFile.split('_')[1].split('.')[0]) + ".bam"
                readSrc = pysam.AlignmentFile(
                    os.path.join(self.inputDirectory, readFile), "r")
                readDest = pysam.AlignmentFile(destFile, "wb",
                                               header=readSrc.header)
                destFilePath = readDest.filename

                for readData in readSrc:
                    readDest.write(readData)
                readDest.close()
                readSrc.close()
                pysam.index(destFilePath)

            # Variants
            print("Converting variants...", file=sys.stderr)
            dsVariantsdir = os.path.join(dsdir, "variants")
            os.makedirs(dsVariantsdir)
            for vgroup in self.datasetVariants[ds].keys():
                vgroupdir = os.path.join(dsVariantsdir, vgroup)
                os.makedirs(vgroupdir)
                for variantFile in self.datasetVariants[ds][vgroup]:
                    destFile = os.path.join(
                        vgroupdir, variantFile.split('_')[2])
                    shutil.copy(
                        os.path.join(
                            self.inputDirectory, variantFile), destFile)
                    # Pysam's tabix_index automatically compresses the file
                    # in place, creates a tabix index.
                    pysam.tabix_index(destFile, preset="vcf")

        print("done converting compliance data.", file=sys.stderr)


@utils.Timed()
def main():
    parser = argparse.ArgumentParser(
        description="Script to generate data bundle from a locally stored "
        "(and possibly locally edited) version of the compliance dataset.")
    parser.add_argument(
        "--outputDirectory", "-o", default="ga4gh-compliance-data",
        help="The directory to output the server-ready data bundle to.")
    parser.add_argument(
        "--inputDirectory", "-i",
        help="Path to local directory containing compliance dataset.",
        default='.')
    parser.add_argument('--verbose', '-v', action='count', default=0)
    args = parser.parse_args()
    cdm = ComplianceDataMunger(args)
    cdm.run()

if __name__ == "__main__":
    main()



"""
Data-driven tests for references.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import glob
import hashlib
import json
import os

# TODO it may be a bit circular to use pysam as our interface for
# accessing reference information, since this is the method we use
# in the main code. We should examine the possibility of using
# a different library here --- perhaps BioPython?
import pysam

import ga4gh.datamodel.references as references
import ga4gh.protocol as protocol
import ga4gh.exceptions as exceptions
import tests.datadriven as datadriven


def testReferenceSets():
    testDataDir = "tests/data/referenceSets"
    for test in datadriven.makeTests(testDataDir, ReferenceSetTest):
        yield test


class ReferenceSetInfo(object):
    """
    Container class for information about a reference set
    """
    def __init__(self, dataDir):
        metadataFilePath = '{}.json'.format(dataDir)
        with open(metadataFilePath) as metadataFile:
            self.metadata = json.load(metadataFile)


class ReferenceInfo(object):
    """
    Container class for information about a reference
    """
    def __init__(self, referenceName, dataDir):
        fastaFileName = os.path.join(
            dataDir, "{}.fa.gz".format(referenceName))
        self.fastaFile = pysam.FastaFile(fastaFileName)
        assert len(self.fastaFile.references) == 1
        assert self.fastaFile.references[0] == referenceName
        self.bases = self.fastaFile.fetch(self.fastaFile.references[0])
        self.length = len(self.bases)
        metadataFileName = os.path.join(
            dataDir, "{}.json".format(referenceName))
        with open(metadataFileName) as metadataFile:
            self.metadata = json.load(metadataFile)


class ReferenceSetTest(datadriven.DataDrivenTest):
    """
    Data drive test class for reference sets. Builds an alternative model of
    a reference set, and verifies that it is consistent with the model
    built by the references.ReferenceSet object.
    """
    def __init__(self, referenceSetId, baseDir):
        super(ReferenceSetTest, self).__init__(referenceSetId, baseDir)
        self._referenceSetInfo = ReferenceSetInfo(baseDir)
        self._referenceInfos = {}
        for fastaFilePath in glob.glob(
                os.path.join(self._dataPath, "*.fa.gz")):
            fastaFileName = os.path.split(fastaFilePath)[1]
            referenceName = fastaFileName.split(".")[0]
            referenceInfo = ReferenceInfo(referenceName, self._dataPath)
            self._referenceInfos[referenceName] = referenceInfo

    def getDataModelInstance(self, localId, dataPath):
        return references.HtslibReferenceSet(localId, dataPath, None)

    def getProtocolClass(self):
        return protocol.ReferenceSet

    def testValidateObjects(self):
        # test that validation works on reference sets and references
        referenceSet = self._gaObject
        referenceSetPe = referenceSet.toProtocolElement()
        self.assertValid(
            protocol.ReferenceSet, referenceSetPe.toJsonDict())
        self.assertGreater(len(referenceSetPe.referenceIds), 0)
        for gaReference in referenceSet.getReferences():
            reference = gaReference.toProtocolElement().toJsonDict()
            self.assertValid(protocol.Reference, reference)

    def testMetadata(self):
        # Check that the metadata loaded from the JSON file is
        # consistent with returned objects.
        referenceSet = self._gaObject
        metadata = self._referenceSetInfo.metadata
        self.assertEqual(
            referenceSet.getAssemblyId(), metadata['assemblyId'])
        self.assertEqual(
            referenceSet.getDescription(), metadata['description'])
        self.assertEqual(
            referenceSet.getIsDerived(), metadata['isDerived'])
        self.assertEqual(
            referenceSet.getNcbiTaxonId(), metadata['ncbiTaxonId'])
        self.assertEqual(
            referenceSet.getSourceAccessions(), metadata['sourceAccessions'])
        self.assertEqual(
            referenceSet.getSourceUri(), metadata['sourceUri'])
        for reference in referenceSet.getReferences():
            referenceInfo = self._referenceInfos[reference.getLocalId()]
            metadata = referenceInfo.metadata
            self.assertEqual(
                reference.getNcbiTaxonId(), metadata["ncbiTaxonId"])
            self.assertEqual(
                reference.getMd5Checksum(), metadata["md5checksum"])
            self.assertEqual(
                reference.getSourceUri(), metadata["sourceUri"])
            self.assertEqual(
                reference.getIsDerived(), metadata["isDerived"])
            self.assertEqual(
                reference.getSourceDivergence(), metadata["sourceDivergence"])
            self.assertEqual(
                reference.getSourceAccessions(), metadata["sourceAccessions"])

    def testGetBases(self):
        # test searching with no arguments succeeds
        referenceSet = self._gaObject
        for gaReference in referenceSet.getReferences():
            pysamReference = self._referenceInfos[gaReference.getLocalId()]
            self.assertReferencesEqual(gaReference, pysamReference)

    def testGetBasesStart(self):
        # test searching with start only succeeds
        self.doRangeTest(5, None)

    def testGetBasesEnd(self):
        # test searching with end only succeeds
        self.doRangeTest(None, 5)

    def testGetBasesRanges(self):
        # test searching with start and end succeeds
        self.doRangeTest(2, 5)

    def testGetBasesEmpty(self):
        self.doRangeTest(0, 0)

    def testOutOfBounds(self):
        referenceSet = self._gaObject
        for reference in referenceSet.getReferences():
            length = reference.getLength()
            badRanges = [
                (-1, 1), (0, length + 1), (0, 2**34), (-2**32, 1),
                (1, 0), (length, length - 1),
            ]
            for start, end in badRanges:
                self.assertRaises(
                    exceptions.ReferenceRangeErrorException,
                    reference.getBases, start, end)

    def testMd5checksums(self):
        referenceSet = self._gaObject
        referenceMd5s = []
        for gaReference in referenceSet.getReferences():
            pysamReference = self._referenceInfos[gaReference.getLocalId()]
            basesChecksum = hashlib.md5(pysamReference.bases).hexdigest()
            self.assertEqual(basesChecksum, gaReference.getMd5Checksum())
            referenceMd5s.append(gaReference.getMd5Checksum())
        referenceMd5s.sort()
        checksumsString = ''.join(referenceMd5s)
        md5checksum = hashlib.md5(checksumsString).hexdigest()
        referenceSetMd5 = referenceSet.getMd5Checksum()
        self.assertEqual(md5checksum, referenceSetMd5)

    def doRangeTest(self, start=None, end=None):
        referenceSet = self._gaObject
        for gaReference in referenceSet.getReferences():
            pysamReference = self._referenceInfos[gaReference.getLocalId()]
            self.assertReferencesEqual(
                gaReference, pysamReference, start, end)

    def assertReferencesEqual(
            self, gaReference, pysamReference, start=None, end=None):
        theStart = 0 if start is None else start
        theEnd = pysamReference.length if end is None else end
        gaBases = gaReference.getBases(theStart, theEnd)
        pysamBases = pysamReference.bases[theStart:theEnd]
        self.assertEqual(gaBases, pysamBases)



"""
Tests the datadriven assert methods
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import tests.datadriven as datadriven


class TestAsserts(unittest.TestCase):

    def setUp(self):
        self.testCase = datadriven.TestCase()

    def testAssertEqual(self):
        self.testCase.assertEqual(0, 0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertEqual(0, 1)

    def testAssertNotEqual(self):
        self.testCase.assertNotEqual(0, 1)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertNotEqual(0, 0)

    def testAssertTrue(self):
        self.testCase.assertTrue(True)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertTrue(False)

    def testAssertFalse(self):
        self.testCase.assertFalse(False)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertFalse(True)

    def testAssertIs(self):
        a = object()
        b = object()
        self.testCase.assertIs(a, a)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIs(a, b)

    def testAssertIsNot(self):
        a = object()
        b = object()
        self.testCase.assertIsNot(a, b)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIsNot(a, a)

    def testAssertIsNone(self):
        self.testCase.assertIsNone(None)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIsNone(0)

    def testAssertIsNotNone(self):
        self.testCase.assertIsNotNone(0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIsNotNone(None)

    def testAssertIn(self):
        self.testCase.assertIn(0, [0])
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIn(0, [])

    def testAssertNotIn(self):
        self.testCase.assertNotIn(0, [])
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertNotIn(0, [0])

    def testAssertIsInstance(self):
        self.testCase.assertIsInstance(0, int)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertIsInstance(0, str)

    def testAssertIsNotInstance(self):
        self.testCase.assertNotIsInstance(0, str)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertNotIsInstance(0, int)

    def testAssertGreater(self):
        self.testCase.assertGreater(1, 0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertGreater(0, 0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertGreater(0, 1)

    def testAssertGreaterEqual(self):
        self.testCase.assertGreaterEqual(0, 0)
        self.testCase.assertGreaterEqual(1, 0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertGreaterEqual(0, 1)

    def testAssertLess(self):
        self.testCase.assertLess(0, 1)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertLess(1, 0)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertLess(0, 0)

    def testAssertLessEqual(self):
        self.testCase.assertLessEqual(0, 0)
        self.testCase.assertLessEqual(0, 1)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertLessEqual(1, 0)

    def testAssertRaises(self):
        def func(*args, **kwargs):
            raise AssertionError()
        self.testCase.assertRaises(AssertionError, func, 1, a=2)

    def testAssertAlmostEqual(self):
        self.testCase.assertAlmostEqual(0.12345678, 0.12345679)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertAlmostEqual(0.123456, 0.1234568)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertAlmostEqual(0.12345678, 0.12345679, 10)

    def testAssertNotAlmostEqual(self):
        self.testCase.assertNotAlmostEqual(0.123456, 0.1234568)
        self.testCase.assertNotAlmostEqual(0.12345678, 0.12345679, 10)
        with self.testCase.assertRaises(AssertionError):
            self.testCase.assertNotAlmostEqual(0.12345678, 0.12345679)



"""
Tests the compound ids
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.datamodel as datamodel
import ga4gh.exceptions as exceptions
import ga4gh.datamodel.datasets as datasets
import ga4gh.datamodel.variants as variants
import ga4gh.datamodel.references as references
import ga4gh.datamodel.reads as reads


class ExampleCompoundId(datamodel.CompoundId):
    separator = ';'
    fields = ['foo', 'bar', 'baz']
    containerIds = [('foobar', 1), ('foobarbaz', 2)]


class TestCompoundIds(unittest.TestCase):
    """
    Test the compound ids
    """
    def testBadParse(self):
        for badId in ['a;b', 'a;b;c;d', 'a;b;sd;', ';;;;']:
            obfuscated = datamodel.CompoundId.obfuscate(badId)
            self.assertEqual(
                badId, datamodel.CompoundId.deobfuscate(obfuscated))
            with self.assertRaises(exceptions.ObjectWithIdNotFoundException):
                ExampleCompoundId.parse(badId)
            with self.assertRaises(exceptions.ObjectWithIdNotFoundException):
                ExampleCompoundId.parse(obfuscated)
        for badType in [0, None, []]:
            with self.assertRaises(exceptions.BadIdentifierException):
                ExampleCompoundId.parse(badType)

    def verifyParseFailure(self, idStr, compoundIdClass):
        """
        Verifies that substrings and superstrings of the specified parsing
        ID string correctly raise parse failures.
        """
        # first, check if we really can parse the string
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = compoundIdClass.parse(obfuscated)
        self.assertIsNotNone(cid)
        # Now, check for substrings
        for j in range(len(idStr) - 1):
            badId = idStr[:j]
            obfuscated = datamodel.CompoundId.obfuscate(badId)
            self.assertRaises(
                exceptions.ObjectWithIdNotFoundException,
                compoundIdClass.parse, obfuscated)
        # Adding on an extra field should also provoke a parse error.
        badId = idStr + ":b"
        obfuscated = datamodel.CompoundId.obfuscate(badId)
        self.assertRaises(
            exceptions.ObjectWithIdNotFoundException, compoundIdClass.parse,
            obfuscated)

    def testAttrs(self):
        obfuscated = datamodel.CompoundId.obfuscate('a;b;c')
        compoundId = ExampleCompoundId.parse(obfuscated)
        self.assertEqual(compoundId.foo, 'a')
        self.assertEqual(compoundId.bar, 'b')
        self.assertEqual(compoundId.baz, 'c')
        obfuscated = datamodel.CompoundId.obfuscate("a;b")
        self.assertEqual(compoundId.foobar, obfuscated)
        obfuscated = datamodel.CompoundId.obfuscate("a;b;c")
        self.assertEqual(compoundId.foobarbaz, obfuscated)

    def testInstantiate(self):
        compoundId = ExampleCompoundId(None, "a", "5", "c")
        obfuscated = datamodel.CompoundId.obfuscate("a;5;c")
        compoundIdStr = str(compoundId)
        self.assertEqual(compoundIdStr, obfuscated)
        self.assertEqual(compoundId.__class__, ExampleCompoundId)

    def getDataset(self):
        return datasets.AbstractDataset("dataset")

    def getReferenceSet(self):
        return references.AbstractReferenceSet("referenceSet")

    def getVariantSet(self):
        return variants.AbstractVariantSet(self.getDataset(), "variantSet")

    def getReadGroupSet(self):
        return reads.AbstractReadGroupSet(self.getDataset(), "readGroupSet")

    def getReadGroup(self):
        return reads.AbstractReadGroup(self.getReadGroupSet(), "readGroup")

    def testDataset(self):
        localId = "dataset"
        cid = datamodel.DatasetCompoundId(None, localId)
        self.assertRaises(
            ValueError, datamodel.DatasetCompoundId, None)
        self.assertEqual(cid.dataset, localId)

    def testDatasetParse(self):
        idStr = "a"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.DatasetCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.verifyParseFailure(idStr, datamodel.DatasetCompoundId)

    def testVariantSet(self):
        dataset = self.getDataset()
        localId = "variantSet"
        cid = datamodel.VariantSetCompoundId(dataset.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.VariantCompoundId, dataset.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.variantSet, localId)
        self.assertEqual(cid.datasetId, dataset.getId())

    def testVariantSetParse(self):
        idStr = "a:b"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.VariantSetCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.variantSet, "b")
        self.verifyParseFailure(idStr, datamodel.VariantSetCompoundId)

    def testCallSet(self):
        name = "sampleName"
        variantSet = self.getVariantSet()
        dataset = variantSet.getParentContainer()
        cid = datamodel.CallSetCompoundId(variantSet.getCompoundId(), name)
        self.assertRaises(
            ValueError, datamodel.CallSetCompoundId,
            variantSet.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.variantSet, variantSet.getLocalId())
        self.assertEqual(cid.name, name)
        self.assertEqual(cid.datasetId, dataset.getId())
        self.assertEqual(cid.variantSetId, variantSet.getId())

    def testCallSetParse(self):
        idStr = "a:b:c"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.CallSetCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.variantSet, "b")
        self.assertEqual(cid.name, "c")
        self.verifyParseFailure(idStr, datamodel.CallSetCompoundId)

    def testVariant(self):
        referenceName = "referenceName"
        start = "start"
        md5 = "md5"
        variantSet = self.getVariantSet()
        dataset = variantSet.getParentContainer()
        cid = datamodel.VariantCompoundId(
            variantSet.getCompoundId(), referenceName, start, md5)
        self.assertRaises(
            ValueError, datamodel.VariantCompoundId,
            variantSet.getCompoundId())
        self.assertRaises(
            ValueError, datamodel.VariantCompoundId,
            variantSet.getCompoundId(), referenceName)
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.variantSet, variantSet.getLocalId())
        self.assertEqual(cid.referenceName, referenceName)
        self.assertEqual(cid.start, start)
        self.assertEqual(cid.md5, md5)
        self.assertEqual(cid.datasetId, dataset.getId())
        self.assertEqual(cid.variantSetId, variantSet.getId())

    def testVariantParse(self):
        idStr = "a:b:c:d:e"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.VariantCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.variantSet, "b")
        self.assertEqual(cid.referenceName, "c")
        self.assertEqual(cid.start, "d")
        self.assertEqual(cid.md5, "e")
        self.verifyParseFailure(idStr, datamodel.VariantCompoundId)

    def testReferenceSet(self):
        localId = "referenceSet"
        cid = datamodel.ReferenceSetCompoundId(None, localId)
        self.assertRaises(
            ValueError, datamodel.ReferenceSetCompoundId, None)
        self.assertEqual(cid.referenceSet, localId)

    def testReferenceSetParse(self):
        idStr = "a"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ReferenceSetCompoundId.parse(obfuscated)
        self.assertEqual(cid.referenceSet, "a")
        self.verifyParseFailure(idStr, datamodel.ReferenceSetCompoundId)

    def testReference(self):
        referenceSet = self.getReferenceSet()
        localId = "reference"
        cid = datamodel.ReferenceCompoundId(
            referenceSet.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.ReferenceCompoundId,
            referenceSet.getCompoundId())
        self.assertEqual(cid.referenceSet, referenceSet.getLocalId())
        self.assertEqual(cid.reference, localId)
        self.assertEqual(cid.referenceSetId, referenceSet.getId())

    def testReferenceParse(self):
        idStr = "a:b"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ReferenceCompoundId.parse(obfuscated)
        self.assertEqual(cid.referenceSet, "a")
        self.assertEqual(cid.reference, "b")
        self.verifyParseFailure(idStr, datamodel.ReferenceCompoundId)

    def testReadGroupSet(self):
        dataset = self.getDataset()
        localId = "readGroupSet"
        cid = datamodel.ReadGroupSetCompoundId(
            dataset.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.ReadGroupSetCompoundId,
            dataset.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.readGroupSet, localId)

    def testReadGroupSetParse(self):
        idStr = "a:b"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ReadGroupSetCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.readGroupSet, "b")
        self.verifyParseFailure(idStr, datamodel.ReadGroupSetCompoundId)

    def testReadGroup(self):
        readGroupSet = self.getReadGroupSet()
        dataset = readGroupSet.getParentContainer()
        localId = "readGroup"
        cid = datamodel.ReadGroupCompoundId(
            readGroupSet.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.ReadGroupCompoundId,
            readGroupSet.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.readGroupSet, readGroupSet.getLocalId())
        self.assertEqual(cid.readGroup, localId)
        self.assertEqual(cid.datasetId, dataset.getId())
        self.assertEqual(cid.readGroupSetId, readGroupSet.getId())

    def testReadGroupParse(self):
        idStr = "a:b:c"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ReadGroupCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.readGroupSet, "b")
        self.assertEqual(cid.readGroup, "c")
        self.verifyParseFailure(idStr, datamodel.ReadGroupCompoundId)

    def testReadAlignment(self):
        readGroup = self.getReadGroup()
        readGroupSet = readGroup.getParentContainer()
        dataset = readGroupSet.getParentContainer()
        localId = "alignment"
        cid = datamodel.ReadAlignmentCompoundId(
            readGroup.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.ReadAlignmentCompoundId,
            dataset.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.readGroupSet, readGroupSet.getLocalId())
        self.assertEqual(cid.readGroup, readGroup.getLocalId())
        self.assertEqual(cid.readAlignment, localId)
        self.assertEqual(cid.datasetId, dataset.getId())
        self.assertEqual(cid.readGroupSetId, readGroupSet.getId())
        self.assertEqual(cid.readGroupId, readGroup.getId())

    def testReadAlignmentParse(self):
        idStr = "a:b:c:d"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ReadAlignmentCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.readGroupSet, "b")
        self.assertEqual(cid.readGroup, "c")
        self.assertEqual(cid.readAlignment, "d")
        self.verifyParseFailure(idStr, datamodel.ReadAlignmentCompoundId)

    def testExperiment(self):
        readGroup = self.getReadGroup()
        readGroupSet = readGroup.getParentContainer()
        dataset = readGroupSet.getParentContainer()
        localId = "experiment"
        cid = datamodel.ExperimentCompoundId(
            readGroup.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.ExperimentCompoundId,
            readGroup.getCompoundId())
        self.assertEqual(cid.dataset, dataset.getLocalId())
        self.assertEqual(cid.readGroupSet, readGroupSet.getLocalId())
        self.assertEqual(cid.readGroup, readGroup.getLocalId())
        self.assertEqual(cid.experiment, localId)

    def testExperimentParse(self):
        idStr = "a:b:c:d"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.ExperimentCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.readGroupSet, "b")
        self.assertEqual(cid.readGroup, "c")
        self.assertEqual(cid.experiment, "d")
        self.verifyParseFailure(idStr, datamodel.ExperimentCompoundId)

    def testVariantSetMetadataCompoundId(self):
        varSet = self.getVariantSet()
        dataSet = varSet.getParentContainer()
        localId = "metadata_key"
        cid = datamodel.VariantSetMetadataCompoundId(
            varSet.getCompoundId(), localId)
        self.assertRaises(
            ValueError, datamodel.VariantSetMetadataCompoundId,
            varSet.getCompoundId())
        self.assertEqual(cid.dataset, dataSet.getLocalId())
        self.assertEqual(cid.variantSet, varSet.getLocalId())
        self.assertEqual(cid.key, localId)

    def testVariantSetMetadataCompoundIdParse(self):
        idStr = "a:b:c"
        obfuscated = datamodel.CompoundId.obfuscate(idStr)
        cid = datamodel.VariantSetMetadataCompoundId.parse(obfuscated)
        self.assertEqual(cid.dataset, "a")
        self.assertEqual(cid.variantSet, "b")
        self.assertEqual(cid.key, "c")
        self.verifyParseFailure(idStr, datamodel.VariantSetMetadataCompoundId)



"""
Provides additional functionality based on avro
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random
import string

import ga4gh.protocol as protocol


INT_MIN_VALUE = -(1 << 31)
INT_MAX_VALUE = (1 << 31) - 1
LONG_MIN_VALUE = -(1 << 63)
LONG_MAX_VALUE = (1 << 63) - 1


class AvrotoolsException(Exception):
    """
    Something that went wrong in the avrotools classes
    """


class AvroTool(object):
    """
    Base class for avrotool classes
    """
    def __init__(self, class_):
        self.class_ = class_
        self.assertProtocolSubclass()
        self.schema = class_.schema

    def assertProtocolSubclass(self):
        if not issubclass(self.class_, protocol.ProtocolElement):
            message = "class '{}' is not subclass of ProtocolElement".format(
                self.class_.__name__)
            raise AvrotoolsException(message)


class Validator(AvroTool):
    """
    Provides methods for validating schemas
    """
    def getInvalidFields(self, jsonDict):
        """
        Takes a jsonDict instantiation of class class_.  If the jsonDict is
        valid for that class, return an empty dictionary.  Otherwise,
        return a multilevel dictionary of the invalid data members and only the
        invalid data members.

        This is an expensive function to run and shouldn't be on the main
        data path.
        """
        validator = SchemaValidator(self.class_)
        invalidFields = validator.getInvalidFields(jsonDict)
        return invalidFields


class Creator(AvroTool):
    """
    Provides methods for creating instances of protocol classes
    according to different rules
    """
    def __init__(self, class_):
        self.class_ = class_
        self.assertProtocolSubclass()
        self.schema = class_.schema

    def getTypicalInstance(self):
        """
        Return a typical instance of the class
        """
        return self._getInstance(TypicalInstanceCreator)

    def getRandomInstance(self):
        """
        Return a random instance of the class
        """
        return self._getInstance(RandomInstanceCreator)

    def getInvalidInstance(self):
        """
        Return an invalid instance of the class
        """
        return self._getInstance(InvalidInstanceCreator)

    def getDefaultInstance(self):
        """
        Return an instance of the class with required values set
        """
        return self._getInstance(DefaultInstanceCreator)

    def _getInstance(self, creatorClass):
        creator = creatorClass(self.class_)
        instance = creator.getInstance()
        return instance

    def getInvalidField(self, fieldName):
        """
        Return an invalid value of the attribute with name fieldName
        """
        return self._getField(InvalidInstanceCreator, fieldName)

    def _getField(self, creatorClass, fieldName):
        creator = creatorClass(self.class_)
        value = creator.getFieldValue(fieldName)
        return value


class AvroTypeSwitch(object):
    """
    Provides reusable logic for branching based on avro types
    """
    schemaRequired = set(
        ['record', 'enum', 'map', 'array', 'union',
         'request', 'error', 'error_union', 'fixed'])

    def __init__(self, class_):
        self.class_ = class_
        self.schema = class_.schema

    def getInstance(self):
        """
        Return an instance of the class instantiated according to
        the rules defined by the subclass
        """
        jsonDict = self.handleSchema(self.schema)
        instance = self.class_.fromJsonDict(jsonDict)
        return instance

    def getFieldValue(self, fieldName):
        """
        Return the value of a field of the class instantiated according
        to the rules defined by the subclass
        """
        fieldType = None
        for field in self.schema.fields:
            if field.name == fieldName:
                fieldType = field.type
                break
        value = self.handleSchema(fieldType)
        return value

    def handleSchema(self, schema, extra=None):
        """
        Return a jsonDict representation of an instance of an object
        defined by schema instantiated according to the rules defined by the
        subclass
        """
        switch = {
            'null': self.handleNull,
            'boolean': self.handleBoolean,
            'string': self.handleString,
            'bytes': self.handleBytes,
            'int': self.handleInt,
            'long': self.handleLong,
            'float': self.handleFloat,
            'double': self.handleDouble,
            'fixed': self.handleFixed,
            'enum': self.handleEnum,
            'array': self.handleArray,
            'map': self.handleMap,
            'union': self.handleUnion,
            'error_union': self.handleErrorUnion,
            'record': self.handleRecord,
            'error': self.handleError,
            'request': self.handleRequest,
        }
        handler = switch[schema.type]
        return self.handleSchemaDispatch(schema, handler, extra)

    def handleSchemaDispatch(self, schema, handler, extra):
        """
        Invoke the handler provided by the handler switch in handleSchema.
        Override if there is a need for extra information to be passed
        into the individual handlers to invoke them correctly.
        """
        if schema.type in self.schemaRequired:
            value = handler(schema)
        else:
            value = handler()
        return value

    # the below methods just delegate to other methods that the subclasses
    # implement; there's no need to re-implement them in (most) subclasses

    def handleErrorUnion(self, schema):
        return self.handleUnion(schema)

    def handleError(self, schema):
        return self.handleRecord(schema)

    def handleRequest(self, schema):
        return self.handleRecord(schema)


class SchemaValidator(AvroTypeSwitch):
    """
    Provides methods for schema validation
    """
    sinkValue = 'aRandomStringBecauseWeCanNotUseNone'

    def handleSchemaDispatch(self, schema, handler, extra):
        if schema.type in self.schemaRequired:
            value = handler(schema, extra)
        else:
            value = handler(extra)
        return value

    def getInvalidFields(self, jsonDict):
        fields = self.handleSchema(self.schema, jsonDict)
        if fields == self.sinkValue:
            return {}
        else:
            return fields

    def handleNull(self, datum):
        if datum is not None:
            return datum
        else:
            return self.sinkValue

    def handleBoolean(self, datum):
        if not isinstance(datum, bool):
            return datum
        else:
            return self.sinkValue

    def handleString(self, datum):
        if not isinstance(datum, basestring):
            return datum
        else:
            return self.sinkValue

    def handleBytes(self, datum):
        if not isinstance(datum, str):
            return datum
        else:
            return self.sinkValue

    def handleInt(self, datum):
        if not ((isinstance(datum, int) or isinstance(datum, long)) and
                INT_MIN_VALUE <= datum <= INT_MAX_VALUE):
            return datum
        else:
            return self.sinkValue

    def handleLong(self, datum):
        if not ((isinstance(datum, int) or isinstance(datum, long)) and
                LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE):
            return datum
        else:
            return self.sinkValue

    def handleFloat(self, datum):
        if not (isinstance(datum, int) or isinstance(datum, long) or
                isinstance(datum, float)):
            return datum
        else:
            return self.sinkValue

    def handleDouble(self, datum):
        return self.handleFloat(datum)

    def handleFixed(self, schema, datum):
        if not (isinstance(datum, str) and len(datum) == schema.size):
            return datum
        else:
            return self.sinkValue

    def handleEnum(self, schema, datum):
        if datum not in schema.symbols:
            return datum
        else:
            return self.sinkValue

    def handleArray(self, schema, datum):
        if not isinstance(datum, list):
            return datum
        arr = []
        for dat in datum:
            result = self.handleSchema(schema.items, dat)
            if result != self.sinkValue:
                arr.append(result)
        if len(arr):
            return arr
        else:
            return self.sinkValue

    def handleMap(self, schema, datum):
        if not isinstance(datum, dict):
            return datum
        dic = {}
        for key, value in datum.items():
            if not isinstance(key, basestring):
                dic[key] = value
            result = self.handleSchema(schema.values, value)
            if result != self.sinkValue:
                dic[key] = result
        if len(dic):
            return dic
        else:
            return self.sinkValue

    def handleUnion(self, schema, datum):
        # If a union type is invalid, we can't tell what was the
        # intended type was to be evaluated.  Therefore, we just
        # take an arbitrary non-null type.
        returnVal = datum
        for unionSchema in schema.schemas:
            result = self.handleSchema(unionSchema, datum)
            if result == self.sinkValue:
                return self.sinkValue
            elif unionSchema.type != 'null':
                returnVal = result
        return returnVal

    def handleRecord(self, schema, datum):
        if not isinstance(datum, dict):
            return datum
        dic = {}
        datumKeys = set(datum.keys())
        for field in schema.fields:
            key = field.name
            value = datum.get(key)
            result = self.handleSchema(field.type, value)
            if result != self.sinkValue:
                dic[key] = result
            try:
                datumKeys.remove(key)
            except KeyError:
                # field that the schema defines as in the record is
                # missing from the jsonDict
                raise AvrotoolsException(key)
        if len(datumKeys):
            # field that the schema does not define is present
            # in the jsonDict
            raise AvrotoolsException(datumKeys)
        if len(dic):
            return dic
        else:
            return self.sinkValue

    def handleErrorUnion(self, schema, datum):
        return self.handleUnion(schema, datum)

    def handleError(self, schema, datum):
        return self.handleRecord(schema, datum)

    def handleRequest(self, schema, datum):
        return self.handleRecord(schema, datum)


class RandomInstanceCreator(AvroTypeSwitch):
    """
    Generates random instances and values
    """
    def handleNull(self):
        return None

    def handleBoolean(self):
        return random.choice([True, False])

    def handleString(self, length=10, characters=string.printable):
        return ''.join(
            [random.choice(characters) for _ in range(length)])

    def handleBytes(self):
        return self.handleString()

    def handleInt(self, min_=INT_MIN_VALUE, max_=INT_MAX_VALUE):
        return random.randint(min_, max_)

    def handleLong(self, min_=LONG_MIN_VALUE, max_=LONG_MAX_VALUE):
        return random.randint(min_, max_)

    def handleFloat(self, min_=INT_MIN_VALUE, max_=INT_MAX_VALUE):
        return random.uniform(min_, max_)

    def handleDouble(self, min_=LONG_MIN_VALUE, max_=LONG_MAX_VALUE):
        return random.uniform(min_, max_)

    def handleFixed(self, schema):
        return self.handleString(schema.size)

    def handleEnum(self, schema):
        return random.choice(schema.symbols)

    def handleArray(self, schema, length=10):
        return [self.handleSchema(schema.items) for _ in range(length)]

    def handleMap(self, schema, size=10):
        return dict(
            (self.handleString(), self.handleSchema(schema.values))
            for _ in range(size))

    def handleUnion(self, schema):
        chosenSchema = random.choice(schema.schemas)
        return self.handleSchema(chosenSchema)

    def handleRecord(self, schema):
        return dict(
            (field.name, self.handleSchema(field.type))
            for field in schema.fields)


class TypicalInstanceCreator(AvroTypeSwitch):
    """
    Generates typical instances and values
    """
    def handleNull(self):
        raise AvrotoolsException()

    def handleBoolean(self):
        return True

    def handleString(self):
        return 'aString'

    def handleBytes(self):
        return b'someBytes'

    def handleInt(self):
        return 5

    def handleLong(self):
        return 6

    def handleFloat(self):
        return 7.0

    def handleDouble(self):
        return 8.0

    def handleFixed(self, schema):
        return 'x' * schema.size

    def handleEnum(self, schema):
        return schema.symbols[0]

    def handleArray(self, schema):
        return [self.handleSchema(schema.items) for _ in range(2)]

    def handleMap(self, schema):
        return {'key': self.handleSchema(schema.values)}

    def handleUnion(self, schema):
        # return an instance of the first non-null schema in the union
        for memberSchema in schema.schemas:
            if memberSchema.type != 'null':
                return self.handleSchema(memberSchema)
        raise AvrotoolsException()  # should never happen

    def handleRecord(self, schema):
        record = {}
        for field in schema.fields:
            record[field.name] = self.handleSchema(field.type)
        return record


class InvalidInstanceCreator(AvroTypeSwitch):
    """
    Generates invalid instances and values
    """
    def handleNull(self):
        return 1

    def handleBoolean(self):
        return "invalidBoolean"

    def handleString(self):
        return 2

    def handleBytes(self):
        return 3

    def handleInt(self):
        return "invalidInt"

    def handleLong(self):
        return "invalidLong"

    def handleFloat(self):
        return "invalidFloat"

    def handleDouble(self):
        return "invalidDouble"

    def handleFixed(self, schema):
        return 4

    def handleEnum(self, schema):
        return "invalidEnum"

    def handleArray(self, schema):
        return "invalidArray"

    def handleMap(self, schema):
        return "invalidMap"

    def handleUnion(self, schema):
        validTypes = set()
        for subSchema in schema.schemas:
            validTypes.add(subSchema.type)
        if 'string' not in validTypes:
            return 'invalidString'
        elif 'map' not in validTypes:
            return {}
        elif 'array' not in validTypes:
            return []
        # no union should have this many types not in it...
        raise AvrotoolsException()

    def handleRecord(self, schema):
        return "invalidRecord"


class DefaultInstanceCreator(TypicalInstanceCreator):
    """
    Generates typical instances with only the default fields set
    """
    def getInstance(self):
        defaultInstance = self.class_()
        jsonDict = defaultInstance.toJsonDict()
        for fieldName in defaultInstance.__slots__:
            if fieldName in defaultInstance.requiredFields:
                typicalValue = self.getFieldValue(fieldName)
                jsonDict[fieldName] = typicalValue
        instance = self.class_.fromJsonDict(jsonDict)
        return instance



"""
Unit tests for faulty data sets.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import unittest

import ga4gh.datamodel.datasets as datasets
import ga4gh.datamodel.references as references
import ga4gh.datamodel.variants as variants
import ga4gh.exceptions as exceptions
import ga4gh.protocol as protocol
import ga4gh.backend as backend


class FaultyVariantDataTest(unittest.TestCase):
    """
    Superclass of faulty variant data tests.
    """
    def setUp(self):
        self.testDataDir = "tests/faultydata/variants"
        self.dataset = datasets.AbstractDataset('dataset1')

    def getFullPath(self, localId):
        return os.path.join(self.testDataDir, localId)


class TestVariantSetNoIndexedVcf(FaultyVariantDataTest):
    localIds = ["no_indexed_vcf"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            self.assertRaises(
                exceptions.NotIndexedException,
                variants.HtslibVariantSet, self.dataset, localId, path,
                None)


class TestInconsistentMetaData(FaultyVariantDataTest):
    localIds = ["inconsist_meta"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            with self.assertRaises(exceptions.InconsistentMetaDataException):
                variants.HtslibVariantSet(self.dataset, localId, path, None)


class TestInconsistentCallSetId(FaultyVariantDataTest):
    localIds = ["inconsist_sampleid", "inconsist_sampleid2"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            with self.assertRaises(exceptions.InconsistentCallSetIdException):
                variants.HtslibVariantSet(self.dataset, localId, path, None)


class TestOverlappingVcfVariants(FaultyVariantDataTest):
    localIds = ["overlapping_vcf"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            with self.assertRaises(exceptions.OverlappingVcfException):
                variants.HtslibVariantSet(self.dataset, localId, path, None)


class TestEmptyDirException(FaultyVariantDataTest):
    localIds = ["empty_dir"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            self.assertRaises(
                exceptions.EmptyDirException,
                variants.HtslibVariantSet, self.dataset, localId, path, None)


class TestDuplicateCallSetId(FaultyVariantDataTest):
    """
    THIS SECTION IS CURRENTLY NOT WORKING
    It returns the following error:

    [E::bcf_hdr_add_sample] Duplicated sample name 'S1'
    Aborted (core dumped)

    which is coming from:
    htslib/vcf.c function bcf_hdr_add_sample

    UNABLE TO CAPTURE EXCEPTION
    """
    localIds = ["duplicated_sampleid"]

    @unittest.skipIf(protocol.version.startswith("0.6"), "")
    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            self.assertRaises(
                exceptions.DuplicateCallSetIdException,
                variants.HtslibVariantSet, self.dataset, localId, path,
                None)


class FaultyReferenceDataTest(unittest.TestCase):
    """
    Superclass of faulty reference data tests
    """

    def getFullPath(self, localId):
        testDataDir = "tests/faultydata/references"
        return os.path.join(testDataDir, localId)


class TestTwoReferences(FaultyReferenceDataTest):
    """
    Tests for FASTA files with more than one reference.
    """

    def testInstantiation(self):
        localId = "two_references"
        path = self.getFullPath(localId)
        self.assertRaises(
            exceptions.NotExactlyOneReferenceException,
            references.HtslibReferenceSet, localId, path, None)


class TestInconsistentReferenceName(FaultyReferenceDataTest):
    """
    Tests the case in which we have a reference file with a different
    name to the ID in the fasta file.
    """

    def testInstantiation(self):
        localId = "inconsistent_reference_name"
        path = self.getFullPath(localId)
        self.assertRaises(
            exceptions.InconsistentReferenceNameException,
            references.HtslibReferenceSet, localId, path, None)


class FaultyReferenceSetDataTest(unittest.TestCase):
    """
    Superclass of faulty reference set data tests
    """
    def getFullPath(self, localId):
        testDataDir = "tests/faultydata/references"
        return os.path.join(testDataDir, localId)


class TestNoReferenceSetMetadata(FaultyReferenceSetDataTest):
    """
    Tests an error is thrown with a missing reference set metadata file
    """
    def testNoReferenceSetMetadata(self):
        localId = "no_refset_meta"
        path = self.getFullPath(localId)
        with self.assertRaises(IOError):
            references.HtslibReferenceSet(localId, path, None)


class TestMissingReferenceSetMetadata(FaultyReferenceSetDataTest):
    """
    Tests an error is thrown with a reference set metadata file that
    is missing entries
    """
    def testMissingReferenceSetMetadata(self):
        localId = "missing_refset_meta"
        path = self.getFullPath(localId)
        with self.assertRaises(exceptions.MissingReferenceSetMetadata):
            references.HtslibReferenceSet(localId, path, None)


class TestInvalidReferenceSetMetadata(FaultyReferenceSetDataTest):
    """
    Tests an error is thrown with a reference set metadata file that
    can not be parsed
    """
    def testMissingReferenceSetMetadata(self):
        localId = "invalid_refset_meta"
        path = self.getFullPath(localId)
        with self.assertRaises(ValueError):
            references.HtslibReferenceSet(localId, path, None)


class FaultyDatasetTest(unittest.TestCase):
    """
    Superclass of faulty dataset tests.
    """
    def setUp(self):
        self.testDataDir = "tests/faultydata/datasets"

    def getFullPath(self, localId):
        return os.path.join(self.testDataDir, localId)


class TestBadDatasetMetadata(FaultyDatasetTest):
    """
    Tests that we raise an expcetion if the metadata is not correct.
    """
    def testBadReferenceDatasetMetadata(self):
        localId = "bad_metadata"
        path = self.getFullPath(localId)
        localBackend = backend.EmptyBackend()
        with self.assertRaises(exceptions.MissingDatasetMetadataException):
            datasets.FileSystemDataset(localId, path, localBackend)



import json

from oic.utils.http_util import Response

from provider.authn import make_cls_from_name, AuthnModule


class UserPass(AuthnModule):
    url_endpoint = "/user_pass/verify"

    def __init__(self, db, template_env, template="user_pass.jinja2", **kwargs):
        super(UserPass, self).__init__(None)
        self.template_env = template_env
        self.template = template

        cls = make_cls_from_name(db["class"])
        self.user_db = cls(**db["kwargs"])

        self.kwargs = kwargs
        self.kwargs.setdefault("page_header", "Log in")
        self.kwargs.setdefault("user_label", "Username")
        self.kwargs.setdefault("passwd_label", "Password")
        self.kwargs.setdefault("submit_btn", "Log in")

    def __call__(self, *args, **kwargs):
        template = self.template_env.get_template(self.template)
        return Response(template.render(action=self.url_endpoint,
                                        state=json.dumps(kwargs),
                                        **self.kwargs))

    def verify(self, *args, **kwargs):
        username = kwargs["username"]
        if username in self.user_db and self.user_db[username] == kwargs[
            "password"]:
            return username, True
        else:
            return self.FAILED_AUTHN



"""
Tests the backend response generators
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.backend as backend
import ga4gh.datamodel.reads as reads
import ga4gh.datamodel.variants as variants
import ga4gh.exceptions as exceptions
import ga4gh.protocol as protocol


def generateVariant():
    variant = protocol.Variant()
    return variant


class MockVariantSet(variants.AbstractVariantSet):

    def __init__(self, parentContainer, localId, numVariants):
        super(MockVariantSet, self).__init__(parentContainer, localId)
        self.numVariants = numVariants

    def getVariants(self, referenceName, startPosition, endPosition,
                    variantName=None, callSetIds=None):
        for i in range(self.numVariants):
            yield generateVariant()


class TestVariantsGenerator(unittest.TestCase):
    """
    Tests the logic of variantsGenerator
    """
    def setUp(self):
        self.request = protocol.SearchVariantsRequest()
        self.backend = backend.SimulatedBackend()
        self.dataset = self.backend.getDatasets()[0]

    def testNonexistantVariantSet(self):
        # a request for a variant set that doesn't exist should throw an error
        variantSet = variants.AbstractVariantSet(
            self.dataset, 'notFound')
        self.request.variantSetId = variantSet.getId()
        with self.assertRaises(exceptions.VariantSetNotFoundException):
            self.backend.variantsGenerator(self.request)

    def testVariantSetEmpty(self):
        # a variant set with no variants should return none
        self._initVariantSet(0)
        iterator = self.backend.variantsGenerator(self.request)
        self.assertIsNone(next(iterator, None))

    def testVariantSetOneVariant(self):
        # a variant set with one variant should return it and a null pageToken
        self._initVariantSet(1)
        iterator = self.backend.variantsGenerator(self.request)
        variant, nextPageToken = next(iterator)
        self.assertIsNotNone(variant)
        self.assertIsNone(nextPageToken)
        self.assertIsNone(next(iterator, None))

    def testVariantSetTwoVariants(self):
        # a variant set with two variants should return the first with
        # a non-null pageToken and the second with a null pageToken
        self._initVariantSet(2)
        iterator = self.backend.variantsGenerator(self.request)
        variant, nextPageToken = next(iterator)
        self.assertIsNotNone(variant)
        self.assertIsNotNone(nextPageToken)
        variant, nextPageToken = next(iterator)
        self.assertIsNotNone(variant)
        self.assertIsNone(nextPageToken)
        self.assertIsNone(next(iterator, None))

    def _initVariantSet(self, numVariants):
        variantSet = MockVariantSet(
            self.dataset, "mockvs", numVariants)
        self.dataset.addVariantSet(variantSet)
        self.request.variantSetId = variantSet.getId()


def generateReadAlignment(position=0, sequence='abc'):
    alignment = protocol.ReadAlignment()
    alignment.alignment = protocol.LinearAlignment()
    alignment.alignment.position = protocol.Position()
    alignment.alignment.position.position = position
    alignment.alignedSequence = sequence
    return alignment


class MockReadGroup(reads.AbstractReadGroup):

    def __init__(self, parentContainer, localId, numAlignments):
        super(MockReadGroup, self).__init__(parentContainer, localId)
        self.numAlignments = numAlignments

    def getReadAlignments(self, referenceName=None, referenceId=None,
                          start=None, end=None):
        for i in range(self.numAlignments):
            yield generateReadAlignment(i)


class TestReadsGenerator(unittest.TestCase):
    """
    Tests the logic of readsGenerator
    """
    def setUp(self):
        self.request = protocol.SearchReadsRequest()
        self.backend = backend.SimulatedBackend(numAlignments=0)
        referenceSet = self.backend.getReferenceSetByIndex(0)
        reference = referenceSet.getReferenceByIndex(0)
        self.request.referenceId = reference.getId()
        self.dataset = self.backend.getDatasets()[0]
        self.readGroupSet = self.dataset.getReadGroupSets()[0]

    def testNoReadGroupsNotSupported(self):
        # a request for no read groups should throw an exception
        self.request.readGroupIds = []
        with self.assertRaises(exceptions.NotImplementedException):
            self.backend.readsGenerator(self.request)

    def testMultipleReadGroupsNotSupported(self):
        # a request for multiple read groups should throw an exception
        self.request.readGroupIds = ["1", "2"]
        with self.assertRaises(exceptions.NotImplementedException):
            self.backend.readsGenerator(self.request)

    def testNonexistantReadGroup(self):
        # a request for a readGroup that doesn't exist should throw an error
        readGroup = reads.AbstractReadGroup(self.readGroupSet, 'notFound')
        self.request.readGroupIds = [readGroup.getId()]
        with self.assertRaises(exceptions.ReadGroupNotFoundException):
            self.backend.readsGenerator(self.request)

    def testReadGroupEmpty(self):
        # a readGroup with no reads should return none
        self._initReadGroup(0)
        iterator = self.backend.readsGenerator(self.request)
        self.assertIsNone(next(iterator, None))

    def testReadGroupOneRead(self):
        # a readGroup with one read should return it and a null nextPageToken
        self._initReadGroup(1)
        iterator = self.backend.readsGenerator(self.request)
        alignment, nextPageToken = next(iterator)
        self.assertIsNotNone(alignment)
        self.assertIsNone(nextPageToken)
        self.assertIsNone(next(iterator, None))

    def testReadGroupTwoReads(self):
        # a readGroup with two reads should return the first with
        # a non-null pageToken and the second with a null pageToken
        self._initReadGroup(2)
        iterator = self.backend.readsGenerator(self.request)
        alignment, nextPageToken = next(iterator)
        self.assertIsNotNone(alignment)
        self.assertIsNotNone(nextPageToken)
        alignment, nextPageToken = next(iterator)
        self.assertIsNotNone(alignment)
        self.assertIsNone(nextPageToken)
        self.assertIsNone(next(iterator, None))

    def _initReadGroup(self, numAlignments):
        readGroup = MockReadGroup(
            self.readGroupSet, "mockrg", numAlignments)
        self.readGroupSet.addReadGroup(readGroup)
        self.request.readGroupIds = [readGroup.getId()]


class TestVariantsIntervalIteratorClassMethods(unittest.TestCase):
    """
    Test the variants interval iterator class methods
    """
    def setUp(self):
        self.variant = protocol.Variant()
        self.variant.start = 4
        self.variant.end = 6
        self.intervalIterator = backend.VariantsIntervalIterator

    def testGetVariantStart(self):
        result = self.intervalIterator._getStart(self.variant)
        self.assertEqual(self.variant.start, result)

    def testGetVariantEnd(self):
        result = self.intervalIterator._getEnd(self.variant)
        self.assertEqual(self.variant.end, result)


class TestReadsIntervalIteratorClassMethods(unittest.TestCase):
    """
    Test the variants interval iterator class methods
    """
    def setUp(self):
        self.read = generateReadAlignment(5)
        self.intervalIterator = backend.ReadsIntervalIterator

    def testGetReadStart(self):
        result = self.intervalIterator._getStart(self.read)
        self.assertEqual(self.read.alignment.position.position, result)

    def testGetReadEnd(self):
        result = self.intervalIterator._getEnd(self.read)
        self.assertEqual(
            self.intervalIterator._getStart(self.read) +
            len(self.read.alignedSequence), result)



"""
Client classes for the GA4GH reference implementation.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests
import posixpath
import logging

import ga4gh.protocol as protocol
import ga4gh.exceptions as exceptions


class AbstractClient(object):
    """
    The abstract superclass of GA4GH Client objects.
    """

    def __init__(self, logLevel=0):
        self._pageSize = None
        self._logLevel = logLevel
        self._protocolBytesReceived = 0
        logging.basicConfig()
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logLevel)

    def _deserializeResponse(self, jsonResponseString, protocolResponseClass):
        self._protocolBytesReceived += len(jsonResponseString)
        self._logger.debug("response:{}".format(jsonResponseString))
        if jsonResponseString == '':
            raise exceptions.EmptyResponseException()
        responseObject = protocolResponseClass.fromJsonString(
            jsonResponseString)
        return responseObject

    def _runSearchPageRequest(
            self, protocolRequest, objectName, protocolResponseClass):
        """
        Runs a complete transaction with the server to obtain a single
        page of search results.
        """
        raise NotImplemented()

    def _runSearchRequest(
            self, protocolRequest, objectName, protocolResponseClass):
        """
        Runs the specified request at the specified objectName and instantiates
        an object of the specified class. We yield each object in listAttr.
        If pages of results are present, repeat this process until the
        pageToken is null.
        """
        notDone = True
        while notDone:
            responseObject = self._runSearchPageRequest(
                protocolRequest, objectName, protocolResponseClass)
            valueList = getattr(
                responseObject, protocolResponseClass.getValueListName())
            for extract in valueList:
                yield extract
            notDone = responseObject.nextPageToken is not None
            protocolRequest.pageToken = responseObject.nextPageToken

    def _runListReferenceBasesPageRequest(self, id_, protocolRequest):
        """
        Runs a complete transaction with the server to get a single
        page of results for the specified ListReferenceBasesRequest.
        """
        raise NotImplemented()

    def listReferenceBases(self, id_, start=0, end=None):
        """
        Returns an iterator over the bases from the server in the form
        of consecutive strings. This command does not conform to the
        patterns of the other search and get requests, and is implemented
        differently.
        """
        request = protocol.ListReferenceBasesRequest()
        request.start = start
        request.end = end
        notDone = True
        # TODO We should probably use a StringIO here to make string buffering
        # a bit more efficient.
        basesList = []
        while notDone:
            response = self._runListReferenceBasesPageRequest(id_, request)
            basesList.append(response.sequence)
            notDone = response.nextPageToken is not None
            request.pageToken = response.nextPageToken
        return "".join(basesList)

    def _runGetRequest(self, objectName, protocolResponseClass, id_):
        """
        Requests an object from the server and returns the object of
        type protocolResponseClass that has id id_.
        Used for requests where a single object is the expected response.
        """
        raise NotImplemented()

    def getPageSize(self):
        """
        Returns the suggested maximum size of pages of results returned by
        the server.
        """
        return self._pageSize

    def setPageSize(self, pageSize):
        """
        Sets the requested maximum size of pages of results returned by the
        server to the specified value.
        """
        self._pageSize = pageSize

    def getProtocolBytesReceived(self):
        """
        Returns the total number of protocol bytes received from the server
        by this client.

        :return: The number of bytes consumed by protocol traffic read from
            the server during the lifetime of this client.
        :rtype: int
        """
        return self._protocolBytesReceived

    def getDataset(self, datasetId):
        """
        Returns the Dataset with the specified ID from the server.

        :param str datasetId: The ID of the Dataset of interest.
        :return: The Dataset of interest.
        :rtype: :class:`ga4gh.protocol.Dataset`
        """
        return self._runGetRequest("datasets", protocol.Dataset, datasetId)

    def getReferenceSet(self, referenceSetId):
        """
        Returns the ReferenceSet with the specified ID from the server.

        :param str referenceSetId: The ID of the ReferenceSet of interest.
        :return: The ReferenceSet of interest.
        :rtype: :class:`ga4gh.protocol.ReferenceSet`
        """
        return self._runGetRequest(
            "referencesets", protocol.ReferenceSet, referenceSetId)

    def getReference(self, referenceId):
        """
        Returns the Reference with the specified ID from the server.

        :param str referenceId: The ID of the Reference of interest.
        :return: The Reference of interest.
        :rtype: :class:`ga4gh.protocol.Reference`
        """
        return self._runGetRequest(
            "references", protocol.Reference, referenceId)

    def getReadGroupSet(self, readGroupSetId):
        """
        Returns the ReadGroupSet with the specified ID from the server.

        :param str readGroupSetId: The ID of the ReadGroupSet of interest.
        :return: The ReadGroupSet of interest.
        :rtype: :class:`ga4gh.protocol.ReadGroupSet`
        """
        return self._runGetRequest(
            "readgroupsets", protocol.ReadGroupSet, readGroupSetId)

    def getReadGroup(self, readGroupId):
        """
        Returns the ReadGroup with the specified ID from the server.

        :param str readGroupId: The ID of the ReadGroup of interest.
        :return: The ReadGroup of interest.
        :rtype: :class:`ga4gh.protocol.ReadGroup`
        """
        return self._runGetRequest(
            "readgroups", protocol.ReadGroup, readGroupId)

    def getCallSet(self, callSetId):
        """
        Returns the CallSet with the specified ID from the server.

        :param str callSetId: The ID of the CallSet of interest.
        :return: The CallSet of interest.
        :rtype: :class:`ga4gh.protocol.CallSet`
        """
        return self._runGetRequest("callsets", protocol.CallSet, callSetId)

    def getVariant(self, variantId):
        """
        Returns the Variant with the specified ID from the server.

        :param str variantId: The ID of the Variant of interest.
        :return: The Variant of interest.
        :rtype: :class:`ga4gh.protocol.Variant`
        """
        return self._runGetRequest("variants", protocol.Variant, variantId)

    def getVariantSet(self, variantSetId):
        """
        Returns the VariantSet with the specified ID from the server.

        :param str variantSetId: The ID of the VariantSet of interest.
        :return: The VariantSet of interest.
        :rtype: :class:`ga4gh.protocol.VariantSet`
        """
        return self._runGetRequest(
            "variantsets", protocol.VariantSet, variantSetId)

    def searchVariants(
            self, variantSetId, start=None, end=None, referenceName=None,
            callSetIds=None):
        """
        Returns an iterator over the Variants fulfilling the specified
        conditions from the specified VariantSet.

        :param str variantSetId: The ID of the
            :class:`ga4gh.protocol.VariantSet` of interest.
        :param int start: Required. The beginning of the window (0-based,
            inclusive) for which overlapping variants should be returned.
            Genomic positions are non-negative integers less than reference
            length. Requests spanning the join of circular genomes are
            represented as two requests one on each side of the join
            (position 0).
        :param int end: Required. The end of the window (0-based, exclusive)
            for which overlapping variants should be returned.
        :param str referenceName: The name of the
            :class:`ga4gh.protocol.Reference` we wish to return variants from.
        :param list callSetIds: Only return variant calls which belong to call
            sets with these IDs. If an empty array, returns variants without
            any call objects. If null, returns all variant calls.

        :return: An iterator over the :class:`ga4gh.protocol.Variant` objects
            defined by the query parameters.
        :rtype: iter
        """
        request = protocol.SearchVariantsRequest()
        request.referenceName = referenceName
        request.start = start
        request.end = end
        request.variantSetId = variantSetId
        request.callSetIds = callSetIds
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "variants", protocol.SearchVariantsResponse)

    def searchDatasets(self):
        """
        Returns an iterator over the Datasets on the server.

        :return: An iterator over the :class:`ga4gh.protocol.Dataset`
            objects on the server.
        """
        request = protocol.SearchDatasetsRequest()
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "datasets", protocol.SearchDatasetsResponse)

    def searchVariantSets(self, datasetId):
        """
        Returns an iterator over the VariantSets fulfilling the specified
        conditions from the specified Dataset.

        :param str datasetId: The ID of the :class:`ga4gh.protocol.Dataset`
            of interest.
        :return: An iterator over the :class:`ga4gh.protocol.VariantSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchVariantSetsRequest()
        request.datasetId = datasetId
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "variantsets", protocol.SearchVariantSetsResponse)

    def searchReferenceSets(
            self, accession=None, md5checksum=None, assemblyId=None):
        """
        Returns an iterator over the ReferenceSets fulfilling the specified
        conditions.

        :param str accession: If not null, return the reference sets for which
            the `accession` matches this string (case-sensitive, exact match).
        :param str md5checksum: If not null, return the reference sets for
            which the `md5checksum` matches this string (case-sensitive, exact
            match). See :class:`ga4gh.protocol.ReferenceSet::md5checksum` for
            details.
        :param str assemblyId: If not null, return the reference sets for which
            the `assemblyId` matches this string (case-sensitive, exact match).
        :return: An iterator over the :class:`ga4gh.protocol.ReferenceSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchReferenceSetsRequest()
        request.accession = accession
        request.md5checksum = md5checksum
        request.assemblyId = assemblyId
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "referencesets", protocol.SearchReferenceSetsResponse)

    def searchReferences(
            self, referenceSetId, accession=None, md5checksum=None):
        """
        Returns an iterator over the References fulfilling the specified
        conditions from the specified Dataset.

        :param str referenceSetId: The ReferenceSet to search.
        :param str accession: If not None, return the references for which the
            `accession` matches this string (case-sensitive, exact match).
        :param str md5checksum: If not None, return the references for which
            the `md5checksum` matches this string (case-sensitive, exact
            match).
        :return: An iterator over the :class:`ga4gh.protocol.Reference`
            objects defined by the query parameters.
        """
        request = protocol.SearchReferencesRequest()
        request.referenceSetId = referenceSetId
        request.accession = accession
        request.md5checksum = md5checksum
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "references", protocol.SearchReferencesResponse)

    def searchCallSets(self, variantSetId, name=None):
        """
        Returns an iterator over the CallSets fulfilling the specified
        conditions from the specified VariantSet.

        :param str name: Only CallSets matching the specified name will
            be returned.
        :return: An iterator over the :class:`ga4gh.protocol.CallSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchCallSetsRequest()
        request.variantSetId = variantSetId
        request.name = name
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "callsets", protocol.SearchCallSetsResponse)

    def searchReadGroupSets(self, datasetId, name=None):
        """
        Returns an iterator over the ReadGroupSets fulfilling the specified
        conditions from the specified Dataset.

        :param str name: Only ReadGroupSets matching the specified name
            will be returned.
        :return: An iterator over the :class:`ga4gh.protocol.ReadGroupSet`
            objects defined by the query parameters.
        :rtype: iter
        """
        request = protocol.SearchReadGroupSetsRequest()
        request.datasetId = datasetId
        request.name = name
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "readgroupsets", protocol.SearchReadGroupSetsResponse)

    def searchReads(
            self, readGroupIds, referenceId=None, start=None, end=None):
        """
        Returns an iterator over the Reads fulfilling the specified
        conditions from the specified ReadGroupIds.

        :param str readGroupIds: The IDs of the
            :class:`ga4gh.protocol.ReadGroup` of interest.
        :param str referenceId: The name of the
            :class:`ga4gh.protocol.Reference` we wish to return reads
            mapped to.
        :param int start: The start position (0-based) of this query. If a
            reference is specified, this defaults to 0. Genomic positions are
            non-negative integers less than reference length. Requests spanning
            the join of circular genomes are represented as two requests one on
            each side of the join (position 0).
        :param int end: The end position (0-based, exclusive) of this query.
            If a reference is specified, this defaults to the reference's
            length.
        :return: An iterator over the
            :class:`ga4gh.protocol.ReadAlignment` objects defined by
            the query parameters.
        :rtype: iter
        """
        request = protocol.SearchReadsRequest()
        request.readGroupIds = readGroupIds
        request.referenceId = referenceId
        request.start = start
        request.end = end
        request.pageSize = self._pageSize
        return self._runSearchRequest(
            request, "reads", protocol.SearchReadsResponse)


class HttpClient(AbstractClient):
    """
    The GA4GH HTTP client. This class provides methods corresponding to the
    GA4GH search and object GET methods.

    .. todo:: Add a better description of the role of this class and include
        links to the high-level API documention.

    :param str urlPrefix: The base URL of the GA4GH server we wish to
        communicate with. This should include the 'http' or 'https' prefix.
    :param int logLevel: The amount of debugging information to log using
        the :mod:`logging` module. This is :data:`logging.WARNING` by default.
    :param str authenticationKey: The authentication key provided by the
        server after logging in.
    """

    def __init__(
            self, urlPrefix, logLevel=logging.WARNING, authenticationKey=None):
        super(HttpClient, self).__init__(logLevel)
        self._urlPrefix = urlPrefix
        self._authenticationKey = authenticationKey
        self._session = requests.Session()
        self._setupHttpSession()
        requestsLog = logging.getLogger("requests.packages.urllib3")
        requestsLog.setLevel(logLevel)
        requestsLog.propagate = True

    def _setupHttpSession(self):
        """
        Sets up the common HTTP session parameters used by requests.
        """
        headers = {"Content-type": "application/json"}
        self._session.headers.update(headers)
        # TODO is this unsafe????
        self._session.verify = False

    def _checkResponseStatus(self, response):
        """
        Checks the speficied HTTP response from the requests package and
        raises an exception if a non-200 HTTP code was returned by the
        server.
        """
        if response.status_code != requests.codes.ok:
            self._logger.error("%s %s", response.status_code, response.text)
            raise exceptions.RequestNonSuccessException(
                "Url {0} had status_code {1}".format(
                    response.url, response.status_code))

    def _getHttpParameters(self):
        """
        Returns the basic HTTP parameters we need all requests.
        """
        return {'key': self._authenticationKey}

    def _runSearchPageRequest(
            self, protocolRequest, objectName, protocolResponseClass):
        url = posixpath.join(self._urlPrefix, objectName + '/search')
        data = protocolRequest.toJsonString()
        self._logger.debug("request:{}".format(data))
        response = self._session.post(
            url, params=self._getHttpParameters(), data=data)
        self._checkResponseStatus(response)
        return self._deserializeResponse(response.text, protocolResponseClass)

    def _runGetRequest(self, objectName, protocolResponseClass, id_):
        urlSuffix = "{objectName}/{id}".format(objectName=objectName, id=id_)
        url = posixpath.join(self._urlPrefix, urlSuffix)
        response = self._session.get(url, params=self._getHttpParameters())
        self._checkResponseStatus(response)
        return self._deserializeResponse(response.text, protocolResponseClass)

    def _runListReferenceBasesPageRequest(self, id_, request):
        urlSuffix = "references/{id}/bases".format(id=id_)
        url = posixpath.join(self._urlPrefix, urlSuffix)
        params = self._getHttpParameters()
        params.update(request.toJsonDict())
        response = self._session.get(url, params=params)
        self._checkResponseStatus(response)
        return self._deserializeResponse(
            response.text, protocol.ListReferenceBasesResponse)


class LocalClient(AbstractClient):

    def __init__(self, backend):
        super(LocalClient, self).__init__()
        self._backend = backend
        self._getMethodMap = {
            "datasets": self._backend.runGetDataset,
            "referencesets": self._backend.runGetReferenceSet,
            "references": self._backend.runGetReference,
            "variantsets": self._backend.runGetVariantSet,
            "variants": self._backend.runGetVariant,
            "readgroupsets": self._backend.runGetReadGroupSet,
            "readgroups": self._backend.runGetReadGroup,
        }
        self._searchMethodMap = {
            "datasets": self._backend.runSearchDatasets,
            "referencesets": self._backend.runSearchReferenceSets,
            "references": self._backend.runSearchReferences,
            "variantsets": self._backend.runSearchVariantSets,
            "variants": self._backend.runSearchVariants,
            "readgroupsets": self._backend.runSearchReadGroupSets,
            "reads": self._backend.runSearchReads,
        }

    def _runGetRequest(self, objectName, protocolResponseClass, id_):
        getMethod = self._getMethodMap[objectName]
        responseJson = getMethod(id_)
        return self._deserializeResponse(responseJson, protocolResponseClass)

    def _runSearchPageRequest(
            self, protocolRequest, objectName, protocolResponseClass):
        searchMethod = self._searchMethodMap[objectName]
        responseJson = searchMethod(protocolRequest.toJsonString())
        return self._deserializeResponse(responseJson, protocolResponseClass)

    def _runListReferenceBasesPageRequest(self, id_, request):
        requestArgs = request.toJsonDict()
        # We need to remove end from this dict if it's not specified because
        # of the way we're interacting with Flask and HTTP GET params.
        # TODO: This is a really nasty way of doing things; we really
        # should just have a request object and pass that around instead of an
        # arguments dictionary.
        if request.end is None:
            del requestArgs["end"]
        if request.pageToken is None:
            del requestArgs["pageToken"]
        responseJson = self._backend.runListReferenceBases(id_, requestArgs)
        return self._deserializeResponse(
            responseJson, protocol.ListReferenceBasesResponse)



"""
Provides classes that take protocol requests, send that request to
the server, and write a particular genomics file type with the results.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections

import pysam

import ga4gh.datamodel.reads as reads
import ga4gh.protocol as protocol


class AbstractConverter(object):
    """
    Abstract base class for converter classes
    """
    def __init__(
            self, container, objectIterator, outputFile, binaryOutput):
        self._container = container
        self._objectIterator = objectIterator
        self._outputFile = outputFile
        self._binaryOutput = binaryOutput


##############################################################################
# SAM
##############################################################################


class SamException(Exception):
    """
    Something that went wrong during converting a SAM file
    """


class SamConverter(object):
    """
    Converts a requested range from a GA4GH server into a SAM file.
    """
    def __init__(
            self, client, readGroupId=None, referenceId=None,
            start=None, end=None, outputFileName=None, binaryOutput=False):
        self._client = client
        self._readGroup = self._client.getReadGroup(readGroupId)
        self._reference = self._client.getReference(referenceId)
        self._start = start
        self._end = end
        self._outputFileName = outputFileName
        self._binaryOutput = binaryOutput

    def convert(self):
        header = self._getHeader()
        targetIds = self._getTargetIds(header)
        # pysam can't write to file streams (except for stdout)
        # http://pysam.readthedocs.org/en/latest/usage.html#using-streams
        if self._binaryOutput:
            flags = "wb"
        else:
            flags = "wh"  # h for header
        fileString = "-"
        if self._outputFileName is not None:
            fileString = self._outputFileName
        alignmentFile = pysam.AlignmentFile(fileString, flags, header=header)
        iterator = self._client.searchReads(
            [self._readGroup.id], self._reference.id, self._start, self._end)
        for read in iterator:
            alignedSegment = SamLine.toAlignedSegment(read, targetIds)
            alignmentFile.write(alignedSegment)
        alignmentFile.close()

    def _getHeader(self):
        # Create header information using self._reference
        header = {
            'HD': {'VN': '1.0'},
            'SQ': [{
                'LN': self._reference.length,
                'SN': self._reference.name
            }]
        }
        return header

    def _getTargetIds(self, header):
        # this seems to be how pysam sets the target ids
        targetIds = collections.defaultdict(int)
        targetId = 0
        if 'SQ' in header:
            headerLines = header['SQ']
            for headerLine in headerLines:
                refName = headerLine['SN']
                targetIds[refName] = targetId
                targetId += 1
        return targetIds


class SamLine(object):
    """
    Methods for processing a line in a SAM file
    """
    _encoding = 'utf8'

    # see tables in SAM spec, section 1.5
    _tagReservedFieldPrefixes = set(["X", "Y", "Z", ])
    _tagIntegerFields = set([
        "AM", "AS", "CM", "CP", "FI", "H0", "H1", "H2", "HI", "IH", "MQ",
        "NH", "NM", "OP", "PQ", "SM", "TC", "UQ", ])
    _tagStringFields = set([
        "BC", "BQ", "CC", "CO", "CQ", "CS", "CT", "E2", "FS", "LB", "MC",
        "MD", "OQ", "OC", "PG", "PT", "PU", "QT", "Q2", "R2", "RG", "RT",
        "SA", "U2", ])
    _tagIntegerArrayFields = set(["FZ", ])

    def __init__(self):
        raise SamException("SamLine can't be instantiated")

    @classmethod
    def toAlignedSegment(cls, read, targetIds):
        ret = pysam.AlignedSegment()
        # QNAME
        ret.query_name = read.fragmentName.encode(cls._encoding)
        # SEQ
        ret.query_sequence = read.alignedSequence.encode(cls._encoding)
        # FLAG
        ret.flag = cls.toSamFlag(read)
        # RNAME
        if read.alignment is not None:
            refName = read.alignment.position.referenceName
            ret.reference_id = targetIds[refName]
        # POS
        if read.alignment is None:
            ret.reference_start = 0
        else:
            ret.reference_start = int(read.alignment.position.position)
        # MAPQ
        if read.alignment is not None:
            ret.mapping_quality = read.alignment.mappingQuality
        # CIGAR
        ret.cigar = cls.toCigar(read)
        # RNEXT
        if read.nextMatePosition is None:
            ret.next_reference_id = -1
        else:
            nextRefName = read.nextMatePosition.referenceName
            ret.next_reference_id = targetIds[nextRefName]
        # PNEXT
        if read.nextMatePosition is None:
            ret.next_reference_start = -1
        else:
            ret.next_reference_start = int(read.nextMatePosition.position)
        # TLEN
        ret.template_length = read.fragmentLength
        # QUAL
        ret.query_qualities = read.alignedQuality
        ret.tags = cls.toTags(read)
        return ret

    @classmethod
    def toSamFlag(cls, read):
        # based on algorithm here:
        # https://github.com/googlegenomics/readthedocs/
        # blob/master/docs/source/migrating_tips.rst
        flag = 0
        if read.numberReads == 2:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.READ_PAIRED)
        if read.properPlacement:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.READ_PROPER_PAIR)
        if read.alignment is None:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.READ_UNMAPPED)
        if read.nextMatePosition is None:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.MATE_UNMAPPED)
        if (read.alignment is not None and
                read.alignment.position.strand ==
                protocol.Strand.NEG_STRAND):
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.READ_REVERSE_STRAND)
        if (read.nextMatePosition is not None and
                read.nextMatePosition.strand == protocol.Strand.NEG_STRAND):
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.MATE_REVERSE_STRAND)
        if read.readNumber is None:
            pass
        elif read.readNumber == 0:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.FIRST_IN_PAIR)
        elif read.readNumber == 1:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.SECOND_IN_PAIR)
        else:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.FIRST_IN_PAIR)
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.SECOND_IN_PAIR)
        if read.secondaryAlignment:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.SECONDARY_ALIGNMENT)
        if read.failedVendorQualityChecks:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.FAILED_QUALITY_CHECK)
        if read.duplicateFragment:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.DUPLICATE_READ)
        if read.supplementaryAlignment:
            flag = reads.SamFlags.setFlag(
                flag, reads.SamFlags.SUPPLEMENTARY_ALIGNMENT)
        return flag

    @classmethod
    def toCigar(cls, read):
        cigarTuples = []
        if read.alignment is not None:
            for gaCigarUnit in read.alignment.cigar:
                operation = reads.SamCigar.ga2int(gaCigarUnit.operation)
                length = int(gaCigarUnit.operationLength)
                cigarTuple = (operation, length)
                cigarTuples.append(cigarTuple)
        return tuple(cigarTuples)

    @classmethod
    def _parseTagValue(cls, tag, value):
        if tag[0] in cls._tagReservedFieldPrefixes:
            # user reserved fields... not really sure what to do here
            return value[0].encode(cls._encoding)
        elif tag in cls._tagIntegerFields:
            return int(value[0])
        elif tag in cls._tagStringFields:
            return value[0].encode(cls._encoding)
        elif tag in cls._tagIntegerArrayFields:
            return [int(integerString) for integerString in value]
        else:
            raise SamException("unrecognized tag '{}'".format(tag))

    @classmethod
    def toTags(cls, read):
        tags = []
        for tag, value in read.info.items():
            val = cls._parseTagValue(tag, value)
            tagTuple = (tag, val)
            tags.append(tagTuple)
        retval = tuple(tags)
        return retval


##############################################################################
# VCF
##############################################################################


class VcfException(Exception):
    pass


class VcfConverter(AbstractConverter):
    """
    Converts the Variants represented by a SearchVariantsRequest into
    VCF format using pysam.
    """
    def _writeHeader(self):
        variantSet = self._container
        # TODO convert this into pysam types and write to the output file.
        # For now, just print out some stuff to demonstrate how to get the
        # attributes we have.
        print("ID = ", variantSet.id)
        print("Dataset ID = ", variantSet.datasetId)
        print("Metadata = ")
        for metadata in variantSet.metadata:
            print("\t", metadata)

    def _writeBody(self):
        for variant in self._objectIterator:
            # TODO convert each variant object into pysam objects and write to
            # the output file. For now, just print the first variant and break.
            print(variant)
            break

    def convert(self):
        """
        Run the conversion process.
        """
        # TODO allocate the pysam VCF object which can be used for the
        # conversion process. See the convert method for ga2sam above.
        self._writeHeader()
        self._writeBody()



"""
Tests the cli
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import mock
import unittest

import ga4gh.cli as cli
import ga4gh.protocol as protocol


class TestGa2VcfArguments(unittest.TestCase):
    """
    Tests the ga2vcf cli can parse all arguments it is supposed to
    """
    def testParseArguments(self):
        cliInput = """--key KEY -O vcf --outputFile /dev/null
        --referenceName REFERENCENAME --callSetIds CALL,SET,IDS --start 0
        --end 1 --pageSize 2 BASEURL VARIANTSETID"""
        parser = cli.getGa2VcfParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.key, "KEY")
        self.assertEqual(args.outputFormat, "vcf")
        self.assertEqual(args.outputFile, "/dev/null")
        self.assertEqual(args.referenceName, "REFERENCENAME")
        self.assertEqual(args.callSetIds, "CALL,SET,IDS")
        self.assertEqual(args.start, 0)
        self.assertEqual(args.end, 1)
        self.assertEqual(args.pageSize, 2)
        self.assertEquals(args.baseUrl, "BASEURL")
        self.assertEquals(args.variantSetId, "VARIANTSETID")


class TestGa2SamArguments(unittest.TestCase):
    """
    Tests the ga2sam cli can parse all arguments it is supposed to
    """
    def testParseArguments(self):
        cliInput = """--key KEY --outputFormat sam
        --pageSize 1 --start 2 --end 3 --outputFile OUT.SAM
        --referenceId REFERENCEID BASEURL READGROUPID"""
        parser = cli.getGa2SamParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.key, "KEY")
        self.assertEqual(args.outputFormat, "sam")
        self.assertEqual(args.outputFile, "OUT.SAM")
        self.assertEqual(args.referenceId, "REFERENCEID")
        self.assertEqual(args.start, 2)
        self.assertEqual(args.end, 3)
        self.assertEqual(args.pageSize, 1)
        self.assertEquals(args.baseUrl, "BASEURL")
        self.assertEquals(args.readGroupId, "READGROUPID")


class TestClientArguments(unittest.TestCase):
    """
    Tests the client cli can parse all arguments it is supposed to
    and can initialize the runner in preparation for a request
    """
    def setUp(self):
        self.parser = cli.getClientParser()

    # TODO we need a way to test parse failures. This is tricky because
    # argparse calls sys.exit() on error, which we can't catch directly as
    # an exception. Using mock to intercept this call would at least
    # verify that an error has been raised.

    def testOutputFormat(self):
        # Most of the commands support the output format option.
        cliInput = "variants-search BASEURL --outputFormat=json"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.outputFormat, "json")
        cliInput = "variants-search BASEURL -O text"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.outputFormat, "text")

    def testVariantsSearchArguments(self):
        cliInput = (
            "variants-search --referenceName REFERENCENAME "
            "--callSetIds CALL,SET,IDS --start 0 "
            "--end 1 --pageSize 2 --variantSetId VARIANTSETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.start, 0)
        self.assertEqual(args.end, 1)
        self.assertEqual(args.referenceName, "REFERENCENAME")
        self.assertEqual(args.callSetIds, "CALL,SET,IDS")
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.variantSetId, "VARIANTSETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.runner, cli.SearchVariantsRunner)

    def testVariantSetsSearchArguments(self):
        cliInput = (
            "variantsets-search --pageSize 1 --datasetId DATASETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchVariantSetsRunner)

    def testReferenceSetsSearchArguments(self):
        cliInput = (
            "referencesets-search --pageSize 1 --accession ACCESSION "
            "--md5checksum MD5CHECKSUM --assemblyId ASSEMBLYID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.md5checksum, "MD5CHECKSUM")
        self.assertEqual(args.assemblyId, "ASSEMBLYID")
        self.assertEqual(args.accession, "ACCESSION")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchReferenceSetsRunner)

    def testReferencesSearchArguments(self):
        cliInput = (
            "references-search --pageSize 10 --accession ACCESSION "
            "--md5checksum MD5CHECKSUM BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 10)
        self.assertEqual(args.md5checksum, "MD5CHECKSUM")
        self.assertEqual(args.accession, "ACCESSION")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchReferencesRunner)

    def testReadGroupSetsSearchArguments(self):
        cliInput = (
            "readgroupsets-search --pageSize 1 --datasetId DATASETID "
            "--name NAME BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.name, "NAME")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchReadGroupSetsRunner)

    def testCallSetsSearchArguments(self):
        cliInput = (
            "callsets-search --pageSize 1 --name NAME "
            "--variantSetId VARIANTSETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.variantSetId, "VARIANTSETID")
        self.assertEqual(args.name, "NAME")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchCallSetsRunner)

    def testReadsSearchArguments(self):
        cliInput = (
            "reads-search --pageSize 2 --start 5 --end 10 "
            "--readGroupIds READ,GROUP,IDS --referenceId REFERENCEID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.start, 5)
        self.assertEqual(args.end, 10)
        self.assertEqual(args.readGroupIds, "READ,GROUP,IDS")
        self.assertEqual(args.referenceId, "REFERENCEID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchReadsRunner)

    def testDatasetsSearchArguments(self):
        cliInput = "datasets-search BASEURL"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli.SearchDatasetsRunner)

    def verifyGetArguments(self, command, runnerClass):
        cliInput = "{} BASEURL ID".format(command)
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.id, "ID")
        self.assertEquals(args.runner, runnerClass)

    def testReferenceSetGetArguments(self):
        self.verifyGetArguments(
            "referencesets-get", cli.GetReferenceSetRunner)

    def testReferenceGetArguments(self):
        self.verifyGetArguments(
            "references-get", cli.GetReferenceRunner)

    def testReadGroupSetGetArguments(self):
        self.verifyGetArguments(
            "readgroupsets-get", cli.GetReadGroupSetRunner)

    def testReadGroupGetArguments(self):
        self.verifyGetArguments(
            "readgroups-get", cli.GetReadGroupRunner)

    def testCallSetGetArguments(self):
        self.verifyGetArguments(
            "callsets-get", cli.GetCallsetRunner)

    def testDatasetsGetArguments(self):
        self.verifyGetArguments(
            "datasets-get", cli.GetDatasetRunner)

    def testVariantGetArguments(self):
        self.verifyGetArguments(
            "variants-get", cli.GetVariantRunner)

    def testReferenceBasesListArguments(self):
        cliInput = (
            "references-list-bases BASEURL ID --start 1 --end 2 "
            "--outputFormat fasta")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.id, "ID")
        self.assertEqual(args.start, 1)
        self.assertEqual(args.end, 2)
        self.assertEquals(args.outputFormat, "fasta")
        self.assertEquals(args.runner, cli.ListReferenceBasesRunner)


class TestOutputFormats(unittest.TestCase):
    """
    Tests the different output formats of the cli
    """
    class FakeArgs(object):
        def __init__(self, outputFormat='text'):
            self.outputFormat = outputFormat
            self.id = 'id'
            self.key = 'key'
            self.baseUrl = 'baseUrl'
            self.verbose = 'verbose'

    class FakeObject(protocol.ProtocolElement):

        __slots__ = ['id', 'name']

        def __init__(self):
            self.id = 'id'
            self.name = 'name'

    def _getRunPrintMethodCalls(self, runner):
        printCalls = []
        with mock.patch('__builtin__.print') as printMethod:
            printMethod.side_effect = \
                lambda *args, **kwargs: printCalls.append((args, kwargs))
            runner.run()
        return printCalls

    def testListReferenceBasesFasta(self):
        args = self.FakeArgs('fasta')
        args.start = 1
        args.end = 100
        returnVal = 'AGCT' * 100  # 400 bases
        runner = cli.ListReferenceBasesRunner(args)
        runner._client.listReferenceBases = mock.Mock(
            return_value=returnVal)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(printCalls[0][0][0], '>id:1-100')
        self.assertEqual(len(printCalls), 7)
        self.assertEqual(
            printCalls[-1][0][0],
            returnVal[-50:])  # 50 = 400 % 70

    def testTextOutput(self):
        returnObj = self.FakeObject()
        args = self.FakeArgs()
        runner = cli.AbstractGetRunner(args)
        runner._method = mock.Mock(return_value=returnObj)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(printCalls, [((u'id', u'name'), {'sep': u'\t'})])

    def testJsonOutput(self):
        returnObj = self.FakeObject()
        args = self.FakeArgs('json')
        runner = cli.AbstractGetRunner(args)
        runner._method = mock.Mock(return_value=returnObj)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(json.loads(printCalls[0][0][0])['name'], 'name')



"""
Script to imitate a Travis CI run
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import glob
import shlex
import subprocess

import yaml

import utils


class TravisSimulator(object):

    logStrPrefix = '***'

    yamlFileLocation = '.travis.yml'

    def parseTestCommands(self):
        yamlFile = file(self.yamlFileLocation)
        yamlData = yaml.load(yamlFile)
        return yamlData['script']

    def expandCommand(self, command):
        # subprocess.check_call doesn't expand globs by default, unless
        # shell=True is passed, which is a security hazzard.
        # This gets around that limitation by implementing some shell features
        # which should be enough for our purposes.  See:
        # https://docs.python.org/2/library/subprocess.html
        # #frequently-used-arguments
        if '*' not in command:
            return command
        splits = shlex.split(command)
        expandedSplits = []
        for split in splits:
            if '*' in split:
                files = glob.glob(split)
                expandedSplits.extend(files)
            else:
                expandedSplits.append(split)
        return ' '.join(expandedSplits)

    def runTests(self):
        testCommands = self.parseTestCommands()
        for command in testCommands:
            expandedCommand = self.expandCommand(command)
            self.log('Running: "{0}"'.format(expandedCommand))
            try:
                utils.runCommand(expandedCommand)
            except subprocess.CalledProcessError:
                self.log('ERROR')
                return
        self.log('SUCCESS')

    def log(self, logStr):
        utils.log("{0} {1}".format(self.logStrPrefix, logStr))


if __name__ == '__main__':
    travisSimulator = TravisSimulator()
    travisSimulator.runTests()



"""
Module responsible for translating reference sequence data into GA4GH native
objects.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import hashlib
import json
import os
import random

import pysam

import ga4gh.datamodel as datamodel
import ga4gh.protocol as protocol
import ga4gh.exceptions as exceptions


DEFAULT_REFERENCESET_NAME = "Default"
"""
This is the name used for any reference set referred to in a BAM
file that does not provide the 'AS' tag in the @SQ header.
"""


class AbstractReferenceSet(datamodel.DatamodelObject):
    """
    Class representing ReferenceSets. A ReferenceSet is a set of
    References which typically comprise a reference assembly, such as
    GRCh38.
    """
    compoundIdClass = datamodel.ReferenceSetCompoundId

    def __init__(self, localId):
        super(AbstractReferenceSet, self).__init__(None, localId)
        self._referenceIdMap = {}
        self._referenceNameMap = {}
        self._referenceIds = []
        self._assemblyId = None
        self._description = None
        self._isDerived = False
        self._ncbiTaxonId = None
        self._sourceAccessions = []
        self._sourceUri = None

    def addReference(self, reference):
        """
        Adds the specified reference to this ReferenceSet.
        """
        id_ = reference.getId()
        self._referenceIdMap[id_] = reference
        self._referenceNameMap[reference.getLocalId()] = reference
        self._referenceIds.append(id_)

    def getReferences(self):
        """
        Returns the References in this ReferenceSet.
        """
        return [self._referenceIdMap[id_] for id_ in self._referenceIds]

    def getNumReferences(self):
        """
        Returns the number of references in this ReferenceSet.
        """
        return len(self._referenceIds)

    def getReferenceByIndex(self, index):
        """
        Returns the reference at the specified index in this ReferenceSet.
        """
        return self._referenceIdMap[self._referenceIds[index]]

    def getReferenceByName(self, name):
        """
        Returns the reference with the specified name.
        """
        if name not in self._referenceNameMap:
            raise exceptions.ReferenceNameNotFoundException(name)
        return self._referenceNameMap[name]

    def getReference(self, id_):
        """
        Returns the Reference with the specified ID or raises a
        ReferenceNotFoundException if it does not exist.
        """
        if id_ not in self._referenceIdMap:
            raise exceptions.ReferenceNotFoundException(id_)
        return self._referenceIdMap[id_]

    def getMd5Checksum(self):
        """
        Returns the MD5 checksum for this reference set. This checksum is
        calculated by making a list of `Reference.md5checksum` for all
        `Reference`s in this set. We then sort this list, and take the
        MD5 hash of all the strings concatenated together.
        """
        references = sorted(
            self.getReferences(),
            key=lambda ref: ref.getMd5Checksum())
        checksums = ''.join([ref.getMd5Checksum() for ref in references])
        md5checksum = hashlib.md5(checksums).hexdigest()
        return md5checksum

    def getAssemblyId(self):
        """
        Returns the assembly ID for this reference set.
        This is the public id of this reference set, such as `GRCh37`
        """
        return self._assemblyId

    def getDescription(self):
        """
        Returns the free text description of this reference set.
        """
        return self._description

    def getIsDerived(self):
        """
        Returns True if this ReferenceSet is derived. A ReferenceSet
        may be derived from a source if it contains additional sequences,
        or some of the sequences within it are derived.
        """
        return self._isDerived

    def getSourceAccessions(self):
        """
        Returns the list of source accession strings. These are all known
        corresponding accession IDs in INSDC (GenBank/ENA/DDBJ) ideally
        with a version number, e.g. `NC_000001.11`.
        """
        return self._sourceAccessions

    def getSourceUri(self):
        """
        Returns the sourceURI for this ReferenceSet.
        """
        return self._sourceUri

    def getNcbiTaxonId(self):
        """
        Returns the NCBI Taxon ID for this reference set. This is the
        ID from http://www.ncbi.nlm.nih.gov/taxonomy (e.g. 9606->human)
        indicating the species which this assembly is intended to model.
        Note that contained `Reference`s may specify a different
        `ncbiTaxonId`, as assemblies may contain reference sequences
        which do not belong to the modeled species, e.g.  EBV in a
        human reference genome.
        """
        return self._ncbiTaxonId

    def toProtocolElement(self):
        """
        Returns the GA4GH protocol representation of this ReferenceSet.
        """
        ret = protocol.ReferenceSet()
        ret.assemblyId = self.getAssemblyId()
        ret.description = self.getDescription()
        ret.id = self.getId()
        ret.isDerived = self.getIsDerived()
        ret.md5checksum = self.getMd5Checksum()
        ret.ncbiTaxonId = self.getNcbiTaxonId()
        ret.referenceIds = self._referenceIds
        ret.sourceAccessions = self.getSourceAccessions()
        ret.sourceURI = self.getSourceUri()
        ret.name = self.getLocalId()
        return ret


class AbstractReference(datamodel.DatamodelObject):
    """
    Class representing References. A Reference is a canonical
    assembled contig, intended to act as a reference coordinate space
    for other genomic annotations. A single Reference might represent
    the human chromosome 1, for instance.
    """
    compoundIdClass = datamodel.ReferenceCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractReference, self).__init__(parentContainer, localId)
        self._length = -1
        self._md5checksum = ""
        self._sourceUri = None
        self._sourceAccessions = []
        self._isDerived = False
        self._sourceDivergence = None
        self._ncbiTaxonId = None

    def getLength(self):
        """
        Returns the length of this reference's sequence string.
        """
        return self._length

    def getName(self):
        """
        Returns the name of this reference, e.g., '22'.
        """
        return self.getLocalId()

    def getIsDerived(self):
        """
        Returns True if this Reference is derived. A sequence X is said to be
        derived from source sequence Y, if X and Y are of the same length and
        the per-base sequence divergence at A/C/G/T bases is sufficiently
        small. Two sequences derived from the same official sequence share the
        same coordinates and annotations, and can be replaced with the official
        sequence for certain use cases.
        """
        return self._isDerived

    def getSourceDivergence(self):
        """
        Returns the source divergence for this reference.  The sourceDivergence
        is the fraction of non-indel bases that do not match the
        reference this record was derived from.
        """
        return self._sourceDivergence

    def getSourceAccessions(self):
        """
        Returns the list of source accession strings. These are all known
        corresponding accession IDs in INSDC (GenBank/ENA/DDBJ) ideally
        with a version number, e.g. `NC_000001.11`.
        """
        return self._sourceAccessions

    def getSourceUri(self):
        """
        The URI from which the sequence was obtained. Specifies a FASTA format
        file/string with one name, sequence pair.
        """
        return self._sourceUri

    def getNcbiTaxonId(self):
        """
        Returns the NCBI Taxon ID for this reference. This is the
        ID from http://www.ncbi.nlm.nih.gov/taxonomy (e.g. 9606->human)
        indicating the species which this assembly is intended to model.
        Note that contained `Reference`s may specify a different
        `ncbiTaxonId`, as assemblies may contain reference sequences
        which do not belong to the modeled species, e.g.  EBV in a
        human reference genome.
        """
        return self._ncbiTaxonId

    def getMd5Checksum(self):
        """
        Returns the MD5 checksum uniquely representing this `Reference` as a
        lower-case hexadecimal string, calculated as the MD5 of the upper-case
        sequence excluding all whitespace characters.
        """
        return self._md5checksum

    def toProtocolElement(self):
        """
        Returns the GA4GH protocol representation of this Reference.
        """
        reference = protocol.Reference()
        reference.id = self.getId()
        reference.isDerived = self.getIsDerived()
        reference.length = self.getLength()
        reference.md5checksum = self.getMd5Checksum()
        reference.name = self.getName()
        reference.ncbiTaxonId = self.getNcbiTaxonId()
        reference.sourceAccessions = self.getSourceAccessions()
        reference.sourceDivergence = self.getSourceDivergence()
        reference.sourceURI = self.getSourceUri()
        return reference

    def checkQueryRange(self, start, end):
        """
        Checks to ensure that the query range is valid within this reference.
        If not, raise ReferenceRangeErrorException.
        """
        condition = (
            (start < 0 or end > self.getLength()) or
            start > end)
        if condition:
            raise exceptions.ReferenceRangeErrorException(
                self.getId(), start, end)

    def getBases(self, start, end):
        """
        Returns the string representing the bases of this reference from
        start (inclusive) to end (exclusive).
        """
        raise NotImplemented()

##################################################################
#
# Simulated references
#
##################################################################


class SimulatedReferenceSet(AbstractReferenceSet):
    """
    A simulated referenceSet
    """
    def __init__(self, localId, randomSeed=0, numReferences=1):
        super(SimulatedReferenceSet, self).__init__(localId)
        self._randomSeed = randomSeed
        self._randomGenerator = random.Random()
        self._randomGenerator.seed(self._randomSeed)
        self._description = "Simulated reference set"
        self._assemblyId = str(random.randint(0, 2**32))
        self._isDerived = bool(random.randint(0, 1))
        self._ncbiTaxonId = random.randint(0, 2**16)
        self._sourceAccessions = []
        for i in range(random.randint(1, 3)):
                self._sourceAccessions.append("sim_accession_{}".format(
                    random.randint(1, 2**32)))
        self._sourceUri = "http://example.com/reference.fa"
        for i in range(numReferences):
            referenceSeed = self._randomGenerator.getrandbits(32)
            referenceLocalId = "srs{}".format(i)
            reference = SimulatedReference(
                self, referenceLocalId, referenceSeed)
            self.addReference(reference)


class SimulatedReference(AbstractReference):
    """
    A simulated reference. Stores a random sequence of a given length, and
    generates remaining attributes randomly.
    """

    def __init__(self, parentContainer, localId, randomSeed=0, length=200):
        super(SimulatedReference, self).__init__(parentContainer, localId)
        rng = random.Random()
        rng.seed(randomSeed)
        self._length = length
        bases = [rng.choice('ACGT') for _ in range(self._length)]
        self._bases = ''.join(bases)
        self._md5checksum = hashlib.md5(self._bases).hexdigest()
        self._isDerived = bool(rng.randint(0, 1))
        self._sourceDivergence = 0
        if self._isDerived:
            self._sourceDivergence = rng.uniform(0, 0.1)
        self._ncbiTaxonId = random.randint(0, 2**16)
        self._sourceAccessions = []
        for i in range(random.randint(1, 3)):
                self._sourceAccessions.append("sim_accession_{}".format(
                    random.randint(1, 2**32)))
        self._sourceUri = "http://example.com/reference.fa"

    def getBases(self, start, end):
        self.checkQueryRange(start, end)
        return self._bases[start:end]

##################################################################
#
# References based on htslib's FASTA file handling.
#
##################################################################


class HtslibReferenceSet(datamodel.PysamDatamodelMixin, AbstractReferenceSet):
    """
    A referenceSet based on data on a file system
    """
    def __init__(self, localId, dataDir, backend):
        super(HtslibReferenceSet, self).__init__(localId)
        self._dataDir = dataDir
        self._setMetadata()
        self._scanDataFiles(dataDir, ["*.fa.gz"])

    def _setMetadata(self):
        metadataFileName = '{}.json'.format(self._dataDir)
        with open(metadataFileName) as metadataFile:
            metadata = json.load(metadataFile)
            try:
                self._assemblyId = metadata['assemblyId']
                self._description = metadata['description']
                self._isDerived = metadata['isDerived']
                self._ncbiTaxonId = metadata['ncbiTaxonId']
                self._sourceAccessions = metadata['sourceAccessions']
                self._sourceUri = metadata['sourceUri']
            except KeyError as err:
                raise exceptions.MissingReferenceSetMetadata(
                    metadataFileName, str(err))

    def _addDataFile(self, path):
        dirname, filename = os.path.split(path)
        localId = filename.split(".")[0]
        metadataFileName = os.path.join(dirname, "{}.json".format(localId))
        with open(metadataFileName) as metadataFile:
            metadata = json.load(metadataFile)
        reference = HtslibReference(self, localId, path, metadata)
        self.addReference(reference)


class HtslibReference(datamodel.PysamDatamodelMixin, AbstractReference):
    """
    A reference based on data stored in a file on the file system
    """
    def __init__(self, parentContainer, localId, dataFile, metadata):
        super(HtslibReference, self).__init__(parentContainer, localId)
        self._fastaFilePath = dataFile
        fastaFile = self.getFileHandle(dataFile)
        numReferences = len(fastaFile.references)
        if numReferences != 1:
            raise exceptions.NotExactlyOneReferenceException(
                self._fastaFilePath, numReferences)
        if fastaFile.references[0] != localId:
            raise exceptions.InconsistentReferenceNameException(
                self._fastaFilePath)
        self._length = fastaFile.lengths[0]
        try:
            self._md5checksum = metadata["md5checksum"]
            self._sourceUri = metadata["sourceUri"]
            self._ncbiTaxonId = metadata["ncbiTaxonId"]
            self._isDerived = metadata["isDerived"]
            self._sourceDivergence = metadata["sourceDivergence"]
            self._sourceAccessions = metadata["sourceAccessions"]
        except KeyError as err:
            raise exceptions.MissingReferenceMetadata(dataFile, str(err))

    def getFastaFilePath(self):
        """
        Returns the fasta file that this reference is derived from.
        """
        return self._fastaFilePath

    def openFile(self, dataFile):
        return pysam.FastaFile(dataFile)

    def getBases(self, start, end):
        self.checkQueryRange(start, end)
        fastaFile = self.getFileHandle(self._fastaFilePath)
        # TODO we should have some error checking here...
        bases = fastaFile.fetch(self.getLocalId(), start, end)
        return bases
